# -*- coding: utf-8 -*-
"""
Module to prepare the training datasets.
"""

from __future__ import print_function
import logging
import shutil
import math
from pathlib import Path
import pprint
from typing import List, Optional, Tuple, Union
import urllib3

from geofileops import geofile
from geofileops import geofileops
import pandas as pd
import geopandas as gpd
import numpy as np
import owslib
import owslib.wms
import owslib.util
from PIL import Image
import rasterio as rio
import rasterio.features as rio_features
import rasterio.profiles as rio_profiles
import shapely.geometry as sh_geom

from orthoseg.util import ows_util
from orthoseg.util import log_util
from orthoseg.helpers.progress_helper import ProgressHelper

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

class LabelInfo:
    def __init__(self,
            locations_path: Path,
            polygons_path: Path,
            image_layer: str):
        self.locations_path = locations_path
        self.polygons_path = polygons_path
        self.image_layer = image_layer
    def __repr__(self):
       return f"LabelInfo with image_layer: {self.image_layer}, locations_path: {self.locations_path}, polygons_path: {self.polygons_path}"

def prepare_traindatasets(
        label_infos: List[LabelInfo],
        classes: dict,
        image_layers: dict,
        training_dir: Path,
        labelname_column: str,
        image_pixel_x_size: float = 0.25,
        image_pixel_y_size: float = 0.25,
        image_pixel_width: int = 512,
        image_pixel_height: int = 512,
        ssl_verify: Union[bool, str] = True,
        force: bool = False) -> Tuple[Path, int]:
    """
    This function prepares training data for the vector labels provided.

    It will:
        * get orthophoto's from a WMS server
        * create the corresponding label mask for each orthophoto
        
    Returns a tuple with (output_dir, dataversion):
        output_dir: the dir where the traindataset was created/found
        dataversion: a version number for the dataset created/found

    Args
        label_infos (List[LabelInfo]): paths to the files with label polygons 
            and locations to generate images for
        classes (dict): dict with the classes to detect as keys. The values 
            are the following:
                - labelnames: list of labels to use for this class
                - weight:
                - burn_value:
        image_layers (dict):
        training_dir (Path):
        labelname_column (str): the column where the label names are stored in 
            the polygon files
        ssl_verify (bool or str, optional): True to use the default 
            certificate bundle as installed on your system. False disables 
            certificate validation (NOT recommended!). If a path to a 
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed 
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
    """
    ### Init stuff ###
    # Interprete ssl_verify
    auth = None
    if ssl_verify is not None: 
        # If it is a string, make sure it isn't actually a bool 
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == 'true':
                ssl_verify = True
            elif ssl_verify.lower() == 'false':
                ssl_verify = False

        auth = owslib.util.Authentication(verify=ssl_verify)
        if ssl_verify is False:
            urllib3.disable_warnings()
            logger.warn("SSL VERIFICATION IS TURNED OFF!!!")
    
    # Check if the first class is named "background"
    if len(classes) == 0:
        raise Exception("No classes specified")
    elif list(classes)[0].lower() != 'background':
        classes_str = pprint.pformat(classes, sort_dicts=False, width=50)
        raise Exception(f"By convention, the first class (the background) should be called 'background'!\n{classes_str}")

    image_crs_width = math.fabs(image_pixel_width*image_pixel_x_size)   # tile width in units of crs => 500 m
    image_crs_height = math.fabs(image_pixel_height*image_pixel_y_size) # tile height in units of crs => 500 m

    ### Check if the latest version of training data is already ok ### 
    # Determine the current data version based on existing output data dir(s),
    # If dir ends on _TMP_* ignore it, as it (probably) ended with an error.
    output_dirs = training_dir.glob(f"[0-9]*/")
    output_dirs = [output_dir for output_dir in output_dirs if not '_TMP_' in output_dir.name]
    
    reuse_traindata = False
    if len(output_dirs) == 0:
        dataversion_new = 1
    else:
        # Get the output dir with the highest version (=first if sorted desc)
        output_dir_mostrecent = sorted(output_dirs, reverse=True)[0]
        dataversion_mostrecent = int(output_dir_mostrecent.name)
        
        # If none of the input files changed since previous run, reuse dataset
        for label_file in label_infos:
            reuse_traindata = True
            labellocations_output_mostrecent_path = (
                    output_dir_mostrecent / label_file.locations_path.name)
            labeldata_output_mostrecent_path = output_dir_mostrecent / label_file.polygons_path.name
            if(not (labellocations_output_mostrecent_path.exists()
                    and labeldata_output_mostrecent_path.exists()
                    and geofile.cmp(label_file.locations_path, labellocations_output_mostrecent_path)
                    and geofile.cmp(label_file.polygons_path, labeldata_output_mostrecent_path))):
                reuse_traindata = False
                break
        if reuse_traindata == True:
            dataversion_new = dataversion_mostrecent
            logger.info(f"Input label file(s) haven't changed since last prepare_traindatasets, so reuse version {dataversion_new}")
        else:
            dataversion_new = dataversion_mostrecent + 1
            logger.info(f"Input label file(s) changed since last prepare_traindatasets, so create new training data version {dataversion_new}")

            # In this case, first check all input files if they are valid...
            invalid_geom_paths = []
            with log_util.LoggingContext(logging.getLogger('geofileops'), level=logging.WARN):
                for label_file in label_infos:
                    is_valid = geofileops.isvalid(label_file.locations_path, force=True)
                    if is_valid is False:
                        invalid_geom_paths.append(str(label_file.locations_path))
                    is_valid = geofileops.isvalid(label_file.polygons_path, force=True)
                    if is_valid is False:
                        invalid_geom_paths.append(str(label_file.polygons_path))
            if len(invalid_geom_paths) > 0:
                raise Exception(f"Invalid geometries found in: {', '.join(invalid_geom_paths)}")
    
    # Determine the output dir 
    training_dataversion_dir = training_dir / f"{dataversion_new:02d}"

    # If the train data is already ok, just return 
    if reuse_traindata is True:
        return (training_dataversion_dir, dataversion_new)

    ### The input labels have changed: copy to new train version ###
    # Create the training dataset first to a temp dir, so it can be 
    # removed/ignored if an error occurs while creating it.
    output_tmp_dir = create_tmp_dir(training_dir, f"{dataversion_new:02d}", remove_existing=True)
    
    labellocations_gdf = None
    labelpolygons_gdf = None
    logger.info(f"Label info: \n{pprint.pformat(label_infos, indent=4)}")
    for label_file in label_infos:

        # Copy the vector files to the dest dir so we keep knowing which files 
        # were used to create the dataset
        geofile.copy(label_file.locations_path, output_tmp_dir)
        geofile.copy(label_file.polygons_path, output_tmp_dir)

        # Read label data and append to general dataframes
        logger.debug(f"Read label locations from {label_file.locations_path}")
        file_labellocations_gdf = geofile.read_file(label_file.locations_path)
        if file_labellocations_gdf is not None and len(file_labellocations_gdf) > 0:
            file_labellocations_gdf.loc[:, 'filepath'] = str(label_file.locations_path)   # type: ignore
            file_labellocations_gdf.loc[:, 'image_layer'] = label_file.image_layer        # type: ignore
            # Remark: geopandas 0.7.0 drops the fid column internally, so cannot be retrieved
            file_labellocations_gdf.loc[:, 'row_nb_orig'] = file_labellocations_gdf.index # type: ignore
            if labellocations_gdf is None:
                labellocations_gdf = file_labellocations_gdf
            else:
                labellocations_gdf = gpd.GeoDataFrame(
                        pd.concat([labellocations_gdf, file_labellocations_gdf], ignore_index=True),
                        crs=file_labellocations_gdf.crs)
        else:
            logger.warn(f"No label locations data found in {label_file.locations_path}")
        logger.debug(f"Read label data from {label_file.polygons_path}")
        file_labelpolygons_gdf = geofile.read_file(label_file.polygons_path)
        if file_labelpolygons_gdf is not None and len(file_labelpolygons_gdf) > 0:
            file_labelpolygons_gdf.loc[:, 'image_layer'] = label_file.image_layer # type: ignore
            if labelpolygons_gdf is None:
                labelpolygons_gdf = file_labelpolygons_gdf
            else:
                labelpolygons_gdf = gpd.GeoDataFrame(
                        pd.concat([labelpolygons_gdf, file_labelpolygons_gdf], ignore_index=True),
                        crs=file_labelpolygons_gdf.crs)
        else:
            logger.warn(f"No label data found in {label_file.polygons_path}")

    # Check if we ended up with labellocations...
    if labellocations_gdf is None:
        raise Exception("Not any labellocation found in the training data, so stop")

    # Get the crs to use from the input vectors...
    img_crs = None
    try:
        img_crs = labellocations_gdf.crs
    except Exception as ex:
        logger.exception(f"Error getting crs from labellocations, labellocation_gdf.crs: {labellocations_gdf.crs}")
        raise ex
    if img_crs is None:
        raise Exception(f"Error getting crs from labellocations, labellocation_gdf.crs: {labellocations_gdf.crs}")

    # Create list with only the input labels that need to be burned in the mask
    if labelpolygons_gdf is not None and labelname_column in labelpolygons_gdf.columns:
        # If there is a column labelname_column (default='label_name'), use the 
        # burn values specified in the configuration
        labels_to_burn_gdf = labelpolygons_gdf
        labels_to_burn_gdf['burn_value'] = None
        for classname in classes:
            labels_to_burn_gdf.loc[(labels_to_burn_gdf[labelname_column].isin(classes[classname]['labelnames'])),
                                   'burn_value'] = classes[classname]['burn_value']
        
        # If there are burn_values that are not filled out, log + stop!
        invalid_labelnames_gdf = labels_to_burn_gdf.loc[labels_to_burn_gdf['burn_value'].isnull()]
        if len(invalid_labelnames_gdf) > 0:
            raise Exception(f"Unknown labelnames (not in config) were found in {len(invalid_labelnames_gdf)} rows, so stop: {invalid_labelnames_gdf[labelname_column].unique()}")
        
        # Filter away rows that are going to burn 0, as this is useless...
        labels_to_burn_gdf = labels_to_burn_gdf.loc[labels_to_burn_gdf['burn_value'] != 0]

    elif len(classes) == 2:
        # There is no column with label names, but there are only 2 classes (background + subsject), so no problem...
        logger.info(f'Column with label names ({labelname_column}) not found, so use all polygons')
        labels_to_burn_gdf = labelpolygons_gdf
        labels_to_burn_gdf.loc[:, 'burn_value'] = 1 # type: ignore

    else:
        # There is no column with label names, but more than two classes, so stop.
        raise Exception(f"Column {labelname_column} is mandatory in labeldata if multiple classes specified: {classes}")

    # Check if we ended up with label data to burn.
    if labels_to_burn_gdf is None:
        raise Exception("Not any labelpolygon retained to burn in the training data, so stop")

    ### Now create the images/masks for the new train version ###
    # Prepare the image data for the different traindata types.
    for traindata_type in ['train', 'validation', 'test']:
        
        # If traindata exists already... continue
        output_imagedatatype_dir = output_tmp_dir / traindata_type
        output_imagedata_image_dir = output_imagedatatype_dir / 'image'
        output_imagedata_mask_dir = output_imagedatatype_dir / 'mask'

        # Create output dirs...
        for dir in [output_imagedatatype_dir, output_imagedata_mask_dir, output_imagedata_image_dir]:
            if dir and not dir.exists():
                dir.mkdir(parents=True, exist_ok=True)

        try:
            # Get the label locations for this traindata type
            labels_to_use_for_bounds_gdf = (
                    labellocations_gdf[labellocations_gdf['traindata_type'] == traindata_type])
            
            # Loop trough all locations labels to get an image for each of them
            nb_todo = len(labels_to_use_for_bounds_gdf)
            logger.info(f"Get images for {nb_todo} {traindata_type} labels")
            created_images_gdf = gpd.GeoDataFrame()
            created_images_gdf['geometry'] = None
            wms_imagelayer_layersources = {}
            progress = ProgressHelper(message="prepare training images", nb_steps_total=nb_todo)
            for i, label_tuple in enumerate(labels_to_use_for_bounds_gdf.itertuples()):
                
                # TODO: update the polygon if it doesn't match the size of the image...
                # as a start... ideally make sure we use it entirely for cases with
                # no abundance of data
                # Make sure the image requested is of the correct size
                label_geom = label_tuple.geometry
                if label_geom is None:
                    logger.warn(f"No geometry found in file {label_tuple.filepath}, (zero based) row_nb_orig: {label_tuple.row_nb_orig}")
                    continue
                geom_bounds = label_geom.bounds
                xmin = geom_bounds[0]-(geom_bounds[0]%image_pixel_x_size)
                ymin = geom_bounds[1]-(geom_bounds[1]%image_pixel_y_size)
                xmax = xmin + image_crs_width
                ymax = ymin + image_crs_height
                img_bbox = sh_geom.box(xmin, ymin, xmax, ymax)
                image_layer = getattr(label_tuple, 'image_layer')

                # If the wms to be used hasn't been initialised yet
                if image_layer not in wms_imagelayer_layersources:
                    wms_imagelayer_layersources[image_layer] = []
                    for layersource in image_layers[image_layer]['layersources']:
                        wms_service = owslib.wms.WebMapService(
                                url=layersource['wms_server_url'], 
                                version=layersource['wms_version'],
                                auth=auth)
                        wms_imagelayer_layersources[image_layer].append(
                                ows_util.LayerSource(
                                        wms_service=wms_service,
                                        layernames=layersource['layernames'],
                                        layerstyles=layersource['layerstyles'],
                                        bands=layersource['bands'],
                                        random_sleep=layersource['random_sleep']))
                                                
                # Now really get the image
                logger.debug(f"Get image for coordinates {img_bbox.bounds}")
                image_filepath = ows_util.getmap_to_file(
                        layersources=wms_imagelayer_layersources[image_layer],
                        output_dir=output_imagedata_image_dir,
                        crs=img_crs,
                        bbox=img_bbox.bounds,
                        size=(image_pixel_width, image_pixel_height),
                        image_format=ows_util.FORMAT_PNG,
                        #image_format_save=ows_util.FORMAT_TIFF,
                        image_pixels_ignore_border=image_layers[image_layer]['image_pixels_ignore_border'],
                        transparent=False,
                        layername_in_filename=True)

                # Create a mask corresponding with the image file
                # image_filepath can be None if file existed already, so check if not None...
                if image_filepath is not None:
                    # Mask should not be in a lossy format!
                    mask_filepath = Path(str(image_filepath)
                            .replace(str(output_imagedata_image_dir), str(output_imagedata_mask_dir))
                            .replace('.jpg', '.png'))
                    nb_classes = len(classes)
                    # Only keep the labels that are meant for this image layer
                    labels_gdf = (labels_to_burn_gdf.loc[
                                        labels_to_burn_gdf['image_layer'] == image_layer]).copy()
                    # assert to evade pyLance warning
                    assert isinstance(labels_gdf, gpd.GeoDataFrame)
                    if len(labels_gdf) == 0:
                        logger.info(f"No labels to be burned in the masks for image_layer {image_layer}!")
                    _create_mask(
                            input_image_filepath=image_filepath,
                            output_mask_filepath=mask_filepath,
                            labels_to_burn_gdf=labels_gdf,
                            nb_classes=nb_classes,
                            force=force)
                
                # Log the progress and prediction speed
                progress.step()

        except Exception as ex:
            raise ex

    # If everything went fine, rename output_tmp_dir to the final output_dir
    output_tmp_dir.rename(training_dataversion_dir)

    return (training_dataversion_dir, dataversion_new)

def create_tmp_dir(
        parent_dir: Path,
        dir_name: str,
        remove_existing: bool = False) -> Path:
    """
    Helper function to create a 'TMP' dir based on a directory name:
        parent_dir / <dir_name>_TMP_<sequence>
    
    Use: if you want to write data to a directory in a "transactional way", 
    it is the safest to write to a temp dir first, and then rename it to the 
    final name. This way, if a hard crash occurs while writing the data, it 
    is clear that the directory wasn't ready. Additionally, in case of a hard 
    crash, file locks can remain which makes it impossible to remove a 
    directory for a while.

    Args:
        parent_dir (Path): The dir to create the temp dir in.
        dir_name (str): The name of the dir to base the temp dir on.
        remove_existing (bool, optional): If True, existing TMP directories 
            will be removed if possible. Defaults to False.

    Raises:
        Exception: [description]

    Returns:
        Path: [description]
    """
    # The final dir should not exists yet!
    final_dir = parent_dir / dir_name
    if final_dir.exists():
        raise Exception(f"It is not supported to create a TMP dir for an existing dir: {final_dir}")
    
    # Try to delete all existing TMP dir's
    if remove_existing is True:
        existing_tmp_dirs = parent_dir.glob(f"{dir_name}_TMP_*")
        for existing_tmp_dir in existing_tmp_dirs:
            try:
                shutil.rmtree(existing_tmp_dir)
            except:
                tmp_dir = None

    # Create the TMP dir
    tmp_dir = None
    for i in range(100):
        tmp_dir = parent_dir / f"{dir_name}_TMP_{i:02d}"

        if tmp_dir.exists():
            # If it (still) exists try next sequence
            tmp_dir = None
            continue
        else:
            # If it doesn't exist, try to create it
            try:
                tmp_dir.mkdir(parents=True)
                break
            except:
                # If it fails to create, try next sequence
                tmp_dir = None
                continue
    
    # If no output tmp dir could be found/created... stop...
    if tmp_dir is None:
        raise Exception(f"Error creating/replacing TMP dir for {dir_name} in {parent_dir}")

    return tmp_dir

def _create_mask(
        input_image_filepath: Path,
        output_mask_filepath: Path,
        labels_to_burn_gdf: gpd.GeoDataFrame,
        nb_classes: int = 1,
        output_imagecopy_filepath: Optional[Path] = None,
        minimum_pct_labeled: float = 0.0,
        force: bool = False) -> Optional[bool]:

    # If file exists already and force is False... stop.
    if(force is False
       and output_mask_filepath.exists()):
        logger.debug(f"Output file already exist, and force is False, return: {output_mask_filepath}")
        return

    # Create a mask corresponding with the image file
    # First read the properties of the input image to copy them for the output
    logger.debug(f"Create mask to {output_mask_filepath}")
    with rio.open(input_image_filepath) as image_ds:
        image_input_profile = image_ds.profile
        image_transform_affine = image_ds.transform

    # Prepare the file profile for the mask depending on output type
    output_ext_lower = output_mask_filepath.suffix.lower()
    if output_ext_lower == '.tif':
        image_output_profile = rio_profiles.DefaultGTiffProfile(
                count=1, transform=image_transform_affine, crs=image_input_profile['crs'])
    if output_ext_lower == '.png':
        image_output_profile = rio_profiles.Profile(driver='PNG', count=1)
    else:
        raise Exception(f"Unsupported mask extension (should be a lossless format!): {output_ext_lower}")
    image_output_profile.update(
            width=image_input_profile['width'], height=image_input_profile['height'], 
            dtype=rio.uint8)

    # Burn the vectors in a mask
    burn_shapes = [(geom, value) 
            for geom, value in zip(labels_to_burn_gdf.geometry, labels_to_burn_gdf.burn_value) if geom is not None]
    if len(burn_shapes) > 0:
        try:
            mask_arr = rio_features.rasterize(
                    shapes=burn_shapes, transform=image_transform_affine,
                    dtype=rio.uint8, fill=0, 
                    out_shape=(image_output_profile['width'], image_output_profile['height']))
            
        except Exception as ex:
            raise Exception(f"Error creating mask for {image_transform_affine}") from ex
    else:
        mask_arr = np.zeros(
                shape=(image_output_profile['width'], image_output_profile['height']),
                dtype=rio.uint8)

    # Check if the mask meets the requirements to be written...
    if minimum_pct_labeled > 0:
        nb_pixels = np.size(mask_arr, 0) * np.size(mask_arr, 1)
        nb_pixels_data = nb_pixels - np.sum(mask_arr == 0)  #np.count_nonnan(image == NoData_value)
        logger.debug(f"nb_pixels: {nb_pixels}, nb_pixels_data: {nb_pixels_data}, pct data: {nb_pixels_data / nb_pixels}")

        if (nb_pixels_data / nb_pixels < minimum_pct_labeled):
            return False

    # Write the labeled mask as .png (so without transform/crs info)
    im = Image.fromarray(mask_arr)
    im.save(output_mask_filepath)
    
    return True

if __name__ == "__main__":
    raise Exception('Not implemented!')



# -*- coding: utf-8 -*-
"""
High-level API to run a segmentation.
"""

import argparse
import logging
import os
from pathlib import Path
import shlex
import sys
import traceback

#import os
#os.environ["CUDA_VISIBLE_DEVICES"] = "-1" # Disable using GPU
import tensorflow as tf

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.lib import predicter
import orthoseg.model.model_factory as mf
import orthoseg.model.model_helper as mh
from orthoseg.util import log_util

from geofileops import geofileops

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def predict_argstr(argstr):
    args = shlex.split(argstr)
    predict_args(args)

def predict_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument('-c', '--config', type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    predict(config_path=Path(args.config))

def predict(config_path: Path):
    """
    Run a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
    """
    ##### Init #####
    # Load the config and save in a bunch of global variables zo it 
    # is accessible everywhere 
    conf.read_orthoseg_config(config_path)
    
    # Init logging
    log_util.clean_log_dir(
            log_dir=conf.dirs.getpath('log_dir'),
            nb_logfiles_tokeep=conf.logging.getint('nb_logfiles_tokeep'))     
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath('log_dir'), __name__)      
    
    # Log start + send email
    message = f"Start predict for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")    
    email_helper.sendmail(message)
    
    try:
        # Read some config, and check if values are ok
        image_layer = conf.image_layers[conf.predict['image_layer']]
        if image_layer is None:
            raise Exception(f"STOP: image_layer to predict is not specified in config: {image_layer}")
        input_image_dir = conf.dirs.getpath('predict_image_input_dir')
        if not input_image_dir.exists():
            raise Exception(f"STOP: input image dir doesn't exist: {input_image_dir}")

        # TODO: add something to delete old data, predictions???

        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement 
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint('force_model_traindata_id')
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id 
        
        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint('trainparams_id')
        best_model = mh.get_best_model(
                model_dir=conf.dirs.getpath('model_dir'), 
                segment_subject=conf.general['segment_subject'],
                traindata_id=traindata_id,
                trainparams_id=trainparams_id)
        
        # Check if a model was found
        if best_model is None:
            message = f"No model found in model_dir: {conf.dirs.getpath('model_dir')} for traindata_id: {traindata_id}"
            logger.critical(message)
            raise Exception(message)
        else:    
            model_weights_filepath = best_model['filepath']
            logger.info(f"Best model found: {model_weights_filepath}")
        
        # Load the hyperparams of the model
        # TODO: move the hyperparams filename formatting to get_models...
        hyperparams_path = best_model['filepath'].parent / f"{best_model['basefilename']}_hyperparams.json"
        hyperparams = mh.HyperParams(path=hyperparams_path)           
        
        # Prepare output subdir to be used for predictions
        predict_out_subdir = f"{best_model['basefilename']}"
        if trainparams_id > 0:
            predict_out_subdir += f"_{trainparams_id}"
        predict_out_subdir += f"_{best_model['epoch']}"
        
        # Try optimizing model with tensorrt. Not supported on Windows
        model = None
        if os.name != 'nt':
            try:
                # Try import
                from tensorflow.python.compiler.tensorrt import trt_convert as trt

                # Import didn't fail, so optimize model
                logger.info('Tensorrt is available, so use optimized model')
                savedmodel_optim_dir = best_model['filepath'].parent / best_model['filepath'].stem + "_optim"
                if not savedmodel_optim_dir.exists():
                    # If base model not yet in savedmodel format
                    savedmodel_dir = best_model['filepath'].parent / best_model['filepath'].stem
                    if not savedmodel_dir.exists():
                        logger.info(f"SavedModel format not yet available, so load model + weights from {best_model['filepath']}")
                        model = mf.load_model(best_model['filepath'], compile=False)
                        logger.info(f"Now save again as savedmodel to {savedmodel_dir}")
                        tf.saved_model.save(model, str(savedmodel_dir))
                        del model

                    # Now optimize model
                    logger.info(f"Optimize + save model to {savedmodel_optim_dir}")
                    converter = trt.TrtGraphConverterV2(
                            input_saved_model_dir=savedmodel_dir,
                            is_dynamic_op=True,
                            precision_mode='FP16')
                    converter.convert()
                    converter.save(savedmodel_optim_dir)
                
                logger.info(f"Load optimized model + weights from {savedmodel_optim_dir}")
                model = tf.keras.models.load_model(savedmodel_optim_dir)

            except ImportError:
                logger.info('Tensorrt is not available, so load unoptimized model')
        
        # If model isn't loaded yet... load!
        if model is None:
            model = mf.load_model(best_model['filepath'], compile=False)

        # Prepare the model for predicting
        nb_gpu = len(tf.config.experimental.list_physical_devices('GPU'))
        batch_size = conf.predict.getint('batch_size')
        if nb_gpu <= 1:
            model_for_predict = model
            logger.info(f"Predict using single GPU or CPU, with nb_gpu: {nb_gpu}")
        else:
            # If multiple GPU's available, create multi_gpu_model
            try:
                model_for_predict = model
                logger.warn(f"Predict using multiple GPUs NOT IMPLEMENTED AT THE MOMENT")
                
                #logger.info(f"Predict using multiple GPUs: {nb_gpu}, batch size becomes: {batch_size*nb_gpu}")
                #batch_size *= nb_gpu
            except ValueError:
                logger.info("Predict using single GPU or CPU")
                model_for_predict = model

        # Prepare some parameters for gthe postprocessing
        simplify_algorithm = conf.predict.get('simplify_algorithm')
        if simplify_algorithm is not None:
            simplify_algorithm = geofileops.SimplifyAlgorithm[simplify_algorithm]
        prediction_cleanup_params = {
                    "simplify_algorithm": simplify_algorithm,
                    "simplify_tolerance": conf.predict.geteval('simplify_tolerance'),
                    "simplify_lookahead": conf.predict.getint('simplify_lookahead'),
                }

        # Prepare the output dirs/paths
        predict_output_dir = Path(f"{str(conf.dirs.getpath('predict_image_output_basedir'))}_{predict_out_subdir}")
        output_vector_dir = conf.dirs.getpath('output_vector_dir')
        output_vector_name = f"{best_model['basefilename']}_{best_model['epoch']}_{conf.predict['image_layer']}"
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"
        
        # Predict for entire dataset
        predicter.predict_dir(
                model=model_for_predict, # type: ignore
                input_image_dir=input_image_dir,
                output_image_dir=predict_output_dir,
                output_vector_path=output_vector_path,
                classes=hyperparams.architecture.classes,
                prediction_cleanup_params=prediction_cleanup_params,
                border_pixels_to_ignore=conf.predict.getint('image_pixels_overlap'),
                projection_if_missing=image_layer['projection'],
                input_mask_dir=None,
                batch_size=batch_size,
                evaluate_mode=False,
                cancel_filepath=conf.files.getpath('cancel_filepath'))
        
        # Log and send mail
        message = f"Completed predict for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running predict for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}")
        raise Exception(message) from ex

def main():
    try:
        predict_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise

# If the script is ran directly...
if __name__ == '__main__':
    main()
    


#!/usr/bin/python3
# Downloaded from https://github.com/sdushantha/gitdir/blob/master/gitdir/gitdir.py

import re
import os
import urllib.request
import json
from pathlib import Path

def create_url(url):
    """
    From the given url, produce a URL that is compatible with Github's REST API. Can handle blob or tree paths.
    """

    # extract the branch name from the given url (e.g master)
    branch = re.findall(r"/tree/(.*?)/", url)
    api_url = url.replace("https://github.com", "https://api.github.com/repos")
    if len(branch) == 0:
        branch = re.findall(r"/blob/(.*?)/", url)[0]
        download_dirs = re.findall(r"/blob/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/blob/.*?/", "/contents/", api_url)
    else:
        branch = branch[0]
        download_dirs = re.findall(r"/tree/" + branch + r"/(.*)", url)[0]
        api_url = re.sub(r"/tree/.*?/", "/contents/", api_url)

    api_url = api_url + "?ref=" + branch
    return api_url, download_dirs

def download(
        repo_url: str, 
        output_dir: Path):
    """ Downloads the files and directories in repo_url. If flatten is specified, the contents of any and all
     sub-directories will be pulled upwards into the root folder. """

    # generate the url which returns the JSON data
    api_url, download_dirs = create_url(repo_url)

    # To handle file names.
    if len(download_dirs.split(".")) == 0:
        dir_out = output_dir / download_dirs
    else:
        dir_out = output_dir / "/".join(download_dirs.split("/")[0:-1])

    response = urllib.request.urlretrieve(api_url)

    # make a directory with the name which is taken from
    # the actual repo
    os.makedirs(str(dir_out), exist_ok=True)

    # total files count
    total_files = 0

    with open(response[0], "r") as f:
        raw_data = f.read()
        data = json.loads(raw_data)

        # Get the total number of files 
        total_files += len(data)

        # If the data is a file, download it as one.
        if isinstance(data, dict) and data["type"] == "file":
            # download the file
            urllib.request.urlretrieve(data["download_url"], dir_out / data["name"])            
            return total_files

        # Loop over all files/dirs found
        for file in data:
            file_url = file["download_url"]
            path = output_dir / file["path"]
            os.makedirs(path.parent, exist_ok=True)

            # If it is a file, download it, if dir, start recursively
            if file_url is not None:
                urllib.request.urlretrieve(file_url, path)
            else:
                download(file["html_url"], output_dir)

    return total_files


# -*- coding: utf-8 -*-
"""
Module with generic usable utility functions to make some tasks using OWS services easier.
"""

import concurrent.futures as futures
import datetime
import itertools as it
import logging
import math
from pathlib import Path
import random
import time
from typing import List, Optional, Tuple, Union
import urllib3
import warnings

import numpy as np
from geofileops import geofile
import geofileops.util.grid_util
import owslib
import owslib.wms
import owslib.util
import pycron
import pyproj
import rasterio as rio
import rasterio.errors as rio_errors
from rasterio import profiles as rio_profiles
from rasterio import transform as rio_transform
from rasterio import windows as rio_windows
import geopandas as gpd
import shapely.geometry as sh_geom

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
FORMAT_GEOTIFF = 'image/geotiff'
FORMAT_GEOTIFF_EXT = '.tif'
FORMAT_GEOTIFF_EXT_WORLD = '.tfw'

FORMAT_TIFF = 'image/tiff'
FORMAT_TIFF_EXT = '.tif'
FORMAT_TIFF_EXT_WORLD = '.tfw'

FORMAT_JPEG = 'image/jpeg'
FORMAT_JPEG_EXT = '.jpg'
FORMAT_JPEG_EXT_WORLD = '.jgw'

FORMAT_PNG = 'image/png'
FORMAT_PNG_EXT = '.png'
FORMAT_PNG_EXT_WORLD = '.pgw'

# Get a logger...
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def get_images_for_grid(
        layersources: List[dict],
        output_image_dir: Path,
        crs: pyproj.CRS,
        image_gen_bounds: Tuple[float, float, float, float] = None,
        image_gen_roi_filepath: Optional[Path] = None,
        grid_xmin: float = 0.0,
        grid_ymin: float = 0.0,
        image_crs_pixel_x_size: float = 0.25,
        image_crs_pixel_y_size: float = 0.25,
        image_pixel_width: int = 1024,
        image_pixel_height: int = 1024,
        image_pixels_ignore_border: int = 0,
        nb_concurrent_calls: int = 1,
        cron_schedule: str = None,     
        image_format: str = FORMAT_GEOTIFF,
        image_format_save: str = None,
        tiff_compress: str = 'lzw',
        transparent: bool = False,
        pixels_overlap: int = 0,
        column_start: int = 0,
        nb_images_to_skip: int = 0,
        max_nb_images: int = -1,
        ssl_verify: Union[bool, str] = True,
        force: bool = False):
    """
    Loads all images in a grid from a WMS service.

    Args:
        layersources (List[dict]): Layer sources to get images from. Multiple 
            sources can be specified to create a combined image, eg. use band 
            1 of a WMS service with band 2 and 3 of another one.
        output_image_dir (Path): Directory to save the images to.
        crs (pyproj.CRS): The crs of the source and destination images.
        image_gen_bounds (Tuple[float, float, float, float], optional): Bounds 
            of the roi to request/save images for. Defaults to None.
        image_gen_roi_filepath (Optional[Path], optional): File with the roi 
            where images should be requested/saved for. Defaults to None.
        grid_xmin (float, optional): xmin for the grid to be used. 
            Defaults to 0.0.
        grid_ymin (float, optional): ymin for the grid to be used. 
            Defaults to 0.0.
        image_crs_pixel_x_size (float, optional): [description]. Defaults to 0.25.
        image_crs_pixel_y_size (float, optional): [description]. Defaults to 0.25.
        image_pixel_width (int, optional): [description]. Defaults to 1024.
        image_pixel_height (int, optional): [description]. Defaults to 1024.
        image_pixels_ignore_border (int, optional): [description]. Defaults to 0.
        nb_concurrent_calls (int, optional): Number of images to treat in 
            parallel. Will increase the load on the WMS server! Defaults to 1.
        cron_schedule (str, optional): [description]. Defaults to None.
        image_format (str, optional): The image format to get. Defaults to 
            FORMAT_GEOTIFF.
        image_format_save (str, optional): The image format to save to. 
            Defaults to None.
        tiff_compress (str, optional): [description]. Defaults to 'lzw'.
        transparent (bool, optional): [description]. Defaults to False.
        pixels_overlap (int, optional): [description]. Defaults to 0.
        column_start (int, optional): [description]. Defaults to 0.
        nb_images_to_skip (int, optional): [description]. Defaults to 0.
        max_nb_images (int, optional): [description]. Defaults to -1.
        ssl_verify (bool or str, optional): True to use the default 
            certificate bundle as installed on your system. False disables 
            certificate validation (NOT recommended!). If a path to a 
            certificate bundle file (.pem) is passed, this will be used.
            In corporate networks using a proxy server this is often needed 
            to evade CERTIFICATE_VERIFY_FAILED errors. Defaults to True.
        force (bool, optional): [description]. Defaults to False.

    Raises:
        Exception: [description]
    """

    ##### Init #####
    if image_format_save is None:
        image_format_save = image_format

    # Interprete ssl_verify
    auth = None
    if ssl_verify is not None: 
        # If it is a string, make sure it isn't actually a bool 
        if isinstance(ssl_verify, str):
            if ssl_verify.lower() == 'true':
                ssl_verify = True
            elif ssl_verify.lower() == 'false':
                ssl_verify = False

        auth = owslib.util.Authentication(verify=ssl_verify)
        if ssl_verify is False:
            urllib3.disable_warnings()
            logger.warn("SSL VERIFICATION IS TURNED OFF!!!")

    crs_width = math.fabs(image_pixel_width*image_crs_pixel_x_size)   # tile width in units of crs => 500 m
    crs_height = math.fabs(image_pixel_height*image_crs_pixel_y_size) # tile height in units of crs => 500 m
    if cron_schedule is not None and cron_schedule != '':
        logger.info(f"A cron_schedule was specified, so the download will only proceed in the specified time range: {cron_schedule}")

    # Read the region of interest file if provided
    roi_gdf = None
    if image_gen_roi_filepath is not None:
        # Read roi
        logger.info(f"Read + optimize region of interest from {image_gen_roi_filepath}")
        roi_gdf = gpd.read_file(str(image_gen_roi_filepath))
                
        # If the generate_window wasn't specified, calculate the bounds
        # based on the roi (but make sure they fit the grid!)
        if image_gen_bounds is None:
            roi_bounds = roi_gdf.geometry.total_bounds
            image_gen_bounds = (roi_bounds[0]-((roi_bounds[0]-grid_xmin)%crs_width),
                                roi_bounds[1]-((roi_bounds[1]-grid_ymin)%crs_height),
                                roi_bounds[2]+(grid_xmin-((roi_bounds[2]-grid_xmin)%crs_width)),
                                roi_bounds[3]+(grid_ymin-((roi_bounds[3]-grid_ymin)%crs_height)))
            logger.debug(f"roi_bounds: {roi_bounds}, image_gen_bounds: {image_gen_bounds}")
            
    # If there is still no image_gen_bounds, stop.
    if image_gen_bounds is None:
        raise Exception("Either image_gen_bounds or an image_gen_roi_filepath should be specified.")

    # Check if the image_gen_bounds are compatible with the grid...
    if(image_gen_bounds[0]-grid_xmin)%crs_width != 0:
        xmin_new = image_gen_bounds[0] - ((image_gen_bounds[0]-grid_xmin)%crs_width)
        logger.warning(f"xmin {image_gen_bounds[0]} incompatible with grid, {xmin_new} will be used")
        image_gen_bounds = (xmin_new, image_gen_bounds[1], image_gen_bounds[2], image_gen_bounds[3])
    if(image_gen_bounds[1]-grid_ymin)%crs_height != 0:
        ymin_new = image_gen_bounds[1] - ((image_gen_bounds[1]-grid_ymin)%crs_height)
        logger.warning(f"ymin {image_gen_bounds[1]} incompatible with grid, {ymin_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], ymin_new, image_gen_bounds[2], image_gen_bounds[3])
    if(image_gen_bounds[2]-grid_xmin)%crs_width != 0:
        xmax_new = image_gen_bounds[2] + crs_width - ((image_gen_bounds[2]-grid_xmin)%crs_width)
        logger.warning(f"xmax {image_gen_bounds[2]} incompatible with grid, {xmax_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], image_gen_bounds[1], xmax_new, image_gen_bounds[3])
    if(image_gen_bounds[3]-grid_ymin)%crs_height != 0:
        ymax_new = image_gen_bounds[3] + crs_height - ((image_gen_bounds[3]-grid_ymin)%crs_height)
        logger.warning(f"ymax {image_gen_bounds[3]} incompatible with grid, {ymax_new} will be used")
        image_gen_bounds = (image_gen_bounds[0], image_gen_bounds[1], image_gen_bounds[2], ymax_new)

    # Create the output dir if it doesn't exist yet
    output_image_dir.mkdir(parents=True, exist_ok=True)

    # Write the cache image grid to file
    grid_path = output_image_dir / "imagecache_grid.gpkg"
    if not grid_path.exists():
        grid_for_roi_gdf = geofileops.util.grid_util.create_grid3(
                image_gen_bounds, width=crs_width, height=crs_height, crs=crs)
        geofile.to_file(grid_for_roi_gdf, grid_path)
    
    # Calculate width and height...
    dx = math.fabs(image_gen_bounds[0] - image_gen_bounds[2]) # area width in units of crs
    dy = math.fabs(image_gen_bounds[1] - image_gen_bounds[3]) # area height in units of crs
    cols = int(math.ceil(dx / crs_width)) + 1
    rows = int(math.ceil(dy / crs_height)) + 1

    # If an roi is defined, split it using a grid as large objects are small
    if roi_gdf is not None:
        # TODO: support creating a grid in latlon????
        if crs.is_projected:
            
            # Create grid
            grid_for_roi_gdf = geofileops.util.grid_util.create_grid3(
                    image_gen_bounds, width=crs_width, height=crs_height, crs=crs)

            # Create intersection layer between grid and roi
            roi_gdf = gpd.overlay(grid_for_roi_gdf, roi_gdf, how='intersection')
            
            # Explode possible multipolygons to polygons
            #roi_gdf.reset_index(drop=True, inplace=True)
            # assert to evade pyLance warning
            assert isinstance(roi_gdf, gpd.GeoDataFrame)
            roi_gdf = roi_gdf.explode(ignore_index=True)
            #roi_gdf.reset_index(drop=True, inplace=True)

            # Write to file
            grid_for_roi_path = output_image_dir / "grid_for_roi.gpkg"
            if grid_path.exists() is False:
                geofile.to_file(grid_for_roi_gdf, grid_path)
            assert isinstance(roi_gdf, gpd.GeoDataFrame)
            gridded_roi_path = output_image_dir / "gridded_roi.gpkg"
            if gridded_roi_path.exists() is False:
                geofile.to_file(roi_gdf, gridded_roi_path)

    # Inits to start getting images    
    layersources_prepared = []
    for layersource in layersources:
        wms_service = owslib.wms.WebMapService(
                url=layersource['wms_server_url'], 
                version=layersource['wms_version'],
                auth=auth)
        layersources_prepared.append(
                LayerSource(
                        wms_service=wms_service,
                        layernames=layersource['layernames'],
                        layerstyles=layersource['layerstyles'],
                        bands=layersource['bands'],
                        random_sleep=layersource['random_sleep']))
                                
    with futures.ThreadPoolExecutor(nb_concurrent_calls) as pool:

        # Loop through all columns and get the images...
        logger.info("Start loading images")
        start_time = None
        start_time_lastprogress = None
        nb_todo = cols*rows
        nb_processed = 0
        nb_processed_lastprogress = 0
        nb_downloaded = 0
        nb_ignore_in_progress = 0
        bbox_list = []
        size_list = []
        output_filename_list = []
        output_dir_list = []
        for col in range(column_start, cols):
            
            image_xmin = col * crs_width + image_gen_bounds[0]
            image_xmax = (col + 1) * crs_width + image_gen_bounds[0]
    
            # If overlapping images are wanted... increase image bbox
            if pixels_overlap:
                image_xmin = image_xmin-(pixels_overlap*image_crs_pixel_x_size)
                image_xmax = image_xmax+(pixels_overlap*image_crs_pixel_x_size)
                
            # Put all the images of this column in a dir
            if crs.is_projected:
                output_dir = output_image_dir / f"{image_xmin:06.0f}"
            else:
                output_dir = output_image_dir / f"{image_xmin:09.4f}"
            if not output_dir.exists():
                output_dir.mkdir()
                
            logger.debug(f"load_images to {output_image_dir.parent.name}/{output_image_dir.name}, column {col} ({output_dir.name})")
            for row in range(0, rows):

                # If a cron_schedule is specified, check if we should be running
                if cron_schedule is not None and cron_schedule != '':
                    # Sleep till the schedule becomes active
                    first_time = True
                    while not pycron.is_now(cron_schedule):
                        # The first time, log message that we are going to sleep...
                        if first_time is True:
                            logger.info(f"There is a time schedule specified, and we need to sleep: {cron_schedule}")
                            first_time = False
                        time.sleep(60)

                nb_processed += 1

                # Calculate y bounds
                image_ymin = row * crs_height + image_gen_bounds[1]
                image_ymax = (row + 1) * crs_height + image_gen_bounds[1]
    
                # If overlapping images are wanted... increase image bbox
                if pixels_overlap:
                    image_ymin = image_ymin-(pixels_overlap*image_crs_pixel_y_size)
                    image_ymax = image_ymax+(pixels_overlap*image_crs_pixel_y_size)
    
                # Create output filename
                output_filename = _create_filename(
                        crs=crs,
                        bbox=(image_xmin, image_ymin, image_xmax, image_ymax),
                        size=(image_pixel_width+2*pixels_overlap, 
                            image_pixel_height+2*pixels_overlap),
                        image_format=image_format,
                        layername=None)
                output_filepath = output_dir / output_filename

                # Do some checks to know if the image needs to be downloaded 
                image_to_be_skipped = False
                if(nb_images_to_skip > 0   
                   and (nb_processed%nb_images_to_skip) != 0):
                    # If we need to skip images, do so...
                    nb_ignore_in_progress += 1
                    image_to_be_skipped = True
                elif not force and output_filepath.exists() and output_filepath.stat().st_size > 0:
                    # Image exists already
                    nb_ignore_in_progress += 1
                    image_to_be_skipped = True
                    logger.debug("    -> image exists already, so skip")
                elif roi_gdf is not None:
                    # If roi was provided, check first if the current image overlaps
                    image_shape = sh_geom.box(image_xmin, image_ymin, image_xmax, image_ymax)

                    possible_match_indexes = list(roi_gdf.geometry.sindex.query(image_shape))
                    possible_matches_gdf = roi_gdf.iloc[possible_match_indexes] # type: ignore
                    precise_matches_gdf = possible_matches_gdf.loc[possible_matches_gdf.intersects(image_shape)]
                    
                    if len(precise_matches_gdf) == 0:
                        nb_ignore_in_progress += 1
                        logger.debug("    -> image doesn't overlap with roi, so skip")
                        image_to_be_skipped = True

                # If the image doesn't need to be skipped... append the image 
                # info to the batch arrays so they can be treated in 
                # bulk if the batch size is reached
                if image_to_be_skipped is False:
                    bbox_list.append((image_xmin, image_ymin, image_xmax, image_ymax))
                    size_list.append((image_pixel_width+2*pixels_overlap, 
                                     image_pixel_height+2*pixels_overlap))
                    output_filename_list.append(output_filename)
                    output_dir_list.append(output_dir)

                # If the batch size is reached or we are at the last images
                nb_images_in_batch = len(bbox_list)
                if(nb_images_in_batch == nb_concurrent_calls or nb_processed == (nb_todo-1)):
    
                    # Now we are getting to start fetching images... init start_time 
                    if start_time is None or start_time_lastprogress is None:
                        start_time = datetime.datetime.now()
                        nb_processed_lastprogress = nb_processed
                    
                    # Exec in parallel 
                    read_results = pool.map(
                            getmap_to_file,        # Function 
                            it.repeat(layersources_prepared),
                            output_dir_list,
                            it.repeat(crs),
                            bbox_list,
                            size_list,
                            it.repeat(image_format),
                            it.repeat(image_format_save),
                            output_filename_list,
                            it.repeat(transparent),
                            it.repeat(tiff_compress),
                            it.repeat(image_pixels_ignore_border),
                            it.repeat(force))
                            
                    for _ in read_results:
                        nb_downloaded += 1
                    
                    # Progress
                    logger.debug(f"Process image {nb_processed} out of {cols*rows}: {nb_processed/(cols*rows):.2f} %")

                    # Log the progress and prediction speed
                    time_between_progress_s = 60
                    time_passed_s = (datetime.datetime.now()-start_time).total_seconds()
                    if start_time_lastprogress is None:
                        time_passed_lastprogress_s = time_between_progress_s
                    else:    
                        time_passed_lastprogress_s = (datetime.datetime.now()-start_time_lastprogress).total_seconds()
                    if time_passed_s > 0 and time_passed_lastprogress_s >= time_between_progress_s:
                        nb_per_hour = ((nb_processed-nb_ignore_in_progress)/time_passed_s) * 3600
                        nb_per_hour_lastprogress = (nb_processed_lastprogress-nb_processed/time_passed_lastprogress_s) * 3600
                        hours_to_go = (int)((nb_todo - nb_processed)/nb_per_hour)
                        min_to_go = (int)((((nb_todo - nb_processed)/nb_per_hour)%1)*60)
                        progress_message = f"load_images to {output_image_dir.parent.name}/{output_image_dir.name}, {hours_to_go:3d}:{min_to_go:2d} left for {nb_todo-nb_processed} images at {nb_per_hour:0.0f}/h ({nb_per_hour_lastprogress:0.0f}/h last batch), with {nb_ignore_in_progress} skipped"
                        logger.info(progress_message)

                        start_time_lastprogress = datetime.datetime.now()
                        nb_processed_lastprogress = nb_processed
                    
                    # Reset variable for next batch
                    bbox_list = []
                    size_list = []
                    output_filename_list = []
                    output_dir_list = []

                    if max_nb_images > -1 and nb_downloaded >= max_nb_images:
                        return            

class LayerSource:
    def __init__(self, 
            wms_service: Union[owslib.wms.wms111.WebMapService_1_1_1, owslib.wms.wms130.WebMapService_1_3_0],
            layernames: List[str],
            layerstyles: List[str] = None,
            bands: List[int] = None,
            random_sleep: int = 0):
        self.wms_service = wms_service
        self.layernames = layernames
        self.layerstyles = layerstyles
        self.bands = bands
        self.random_sleep = random_sleep

def getmap_to_file(
        layersources: List[LayerSource],
        output_dir: Path,
        crs: Union[str, pyproj.CRS],
        bbox: Tuple[float, float, float, float],
        size: Tuple[int, int],
        image_format: str = FORMAT_GEOTIFF,
        image_format_save: str = None,
        output_filename: str = None,
        transparent: bool = False,
        tiff_compress: str = 'lzw',
        image_pixels_ignore_border: int = 0,
        force: bool = False,
        layername_in_filename: bool = False) -> Optional[Path]:
    """
    

    Args:
        layersources (List[dict]): Layer sources to get images from. Multiple 
            sources can be specified to create a combined image, eg. use band 
            1 of a WMS service with band 2 and 3 of another one.
        output_image_dir (Path): Directory to save the images to.
        crs (pyproj.CRS): The crs of the source and destination images.
        bbox (Tuple[float, float, float, float]): Bbox of the image to get.
        size (Tuple[int, int]): The image width and height.
        image_format (str, optional): [description]. Defaults to FORMAT_GEOTIFF.
        image_format_save (str, optional): [description]. Defaults to None.
        output_filename (str, optional): [description]. Defaults to None.
        transparent (bool, optional): [description]. Defaults to False.
        tiff_compress (str, optional): [description]. Defaults to 'lzw'.
        image_pixels_ignore_border (int, optional): [description]. Defaults to 0.
        force (bool, optional): [description]. Defaults to False.
        layername_in_filename (bool, optional): [description]. Defaults to False.

    Raises:
        Exception: [description]
        Exception: [description]
        Exception: [description]
        Exception: [description]

    Returns:
        Optional[Path]: [description]
    """

    ##### Init #####
    # If no seperate save format is specified, use the standard image_format
    if image_format_save is None:
        image_format_save = image_format

    # If crs is specified as str, convert to CRS
    if isinstance(crs, str):
        crs = pyproj.CRS(crs)

    # If there isn't a filename supplied, create one...
    if output_filename is None:
        layername = None
        if layername_in_filename:
            for layersource in layersources:
                if layername is None:
                    layername = '_'.join(layersource.layernames)
                else:
                    layername += f"_{'_'.join(layersource.layernames)}"

        output_filename = _create_filename(
                crs=crs, 
                bbox=bbox, size=size, 
                image_format=image_format_save,
                layername=layername)

    # Create full output filepath
    output_filepath = output_dir / output_filename

    # If force is false and file exists already, stop...
    if force is False and output_filepath.exists(): 
        if output_filepath.stat().st_size > 0:
            logger.debug(f"File already exists, skip: {output_filepath}")
            return None
        else:
            output_filepath.unlink()

    logger.debug(f"Get image to {output_filepath}")

    if not output_dir.exists():
        output_dir.mkdir()

    ##### Get image(s), read the band to keep and save #####
    # Some hacks for special cases...
    bbox_for_getmap = bbox
    size_for_getmap = size
    # Dirty hack to ask a bigger picture, and then remove the border again!
    if image_pixels_ignore_border > 0:
        x_pixsize = (bbox[2]-bbox[0])/size[0]
        y_pixsize = (bbox[3]-bbox[1])/size[1]
        bbox_for_getmap = (bbox[0] - x_pixsize*image_pixels_ignore_border,
                        bbox[1] - y_pixsize*image_pixels_ignore_border,
                        bbox[2] + x_pixsize*image_pixels_ignore_border,
                        bbox[3] + y_pixsize*image_pixels_ignore_border)
        size_for_getmap = (size[0] + 2*image_pixels_ignore_border,
                        size[1] + 2*image_pixels_ignore_border)
    # Dirty hack to support y,x cordinate system
    if crs.to_epsg() == 3059:
        bbox_for_getmap = (bbox_for_getmap[1], bbox_for_getmap[0], 
                        bbox_for_getmap[3], bbox_for_getmap[2])

    image_data_output = None
    image_profile_output = None
    response = None
    for layersource in layersources:
  
        # Get image from server, and retry up to 10 times...
        nb_retries = 0
        time_sleep = 5
        image_retrieved = False
        while image_retrieved is False:
            
            try:
                logger.debug(f"Start call GetMap for bbox {bbox}")
                response = layersource.wms_service.getmap(
                        layers=layersource.layernames,
                        styles=layersource.layerstyles,
                        srs=f"epsg:{crs.to_epsg()}",
                        bbox=bbox_for_getmap,
                        size=size_for_getmap,
                        format=image_format,
                        transparent=transparent)
                logger.debug(f"Finished doing request {response.geturl()}")
                
                # If a random sleep was specified... apply it
                if layersource.random_sleep > 0:
                    time.sleep(random.uniform(0, layersource.random_sleep))
                
                # Image was retrieved... so stop loop
                image_retrieved = True
            except Exception as ex:
                if isinstance(ex, owslib.util.ServiceException):
                    if 'Error rendering coverage on the fast path' in str(ex):
                        logger.error(f"Request for bbox {bbox_for_getmap} gave a non-blocking exception, SKIP and proceed: {ex}") 
                        return
                    elif 'java.lang.OutOfMemoryError: Java heap space' in str(ex):
                        logger.debug(f"Request for bbox {bbox_for_getmap} gave a non-blocking exception, try again in {time_sleep} s: {ex}") 
                    else:
                        raise Exception(f"WMS Service gave an exception for bbox {bbox_for_getmap}: {ex}") from ex
                
                # If the exception isn't handled yet, retry 10 times... 
                if nb_retries < 10:
                    time.sleep(time_sleep)
                    
                    # Increase sleep time every time.
                    time_sleep += 5
                    nb_retries += 1
                    continue
                else:
                    message = f"Retried 10 times and didn't work, with layers: {layersource.layernames}, styles: {layersource.layernames}, for bbox: {bbox_for_getmap}"
                    logger.exception(message)
                    raise Exception(message) from ex

        # If the response is None, error
        if response is None:
            raise Exception("No valid response retrieved...")
        
        # Write image to temp file...
        # If all bands need to be kept, just save to output
        with rio.MemoryFile(response.read()) as memfile:

            # Because the image returned by WMS doesn't contain georeferencing 
            # info, suppress NotGeoreferencedWarning
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=rio_errors.NotGeoreferencedWarning)
    
                with memfile.open() as image_ds:
                    image_profile_curr = image_ds.profile

                    # Read the data we need from the memoryfile 
                    if layersource.bands is None:
                        # If no specific bands specified, read them all...
                        if image_data_output is None:
                            image_data_output = image_ds.read()
                        else:
                            image_data_output = np.append(image_data_output, image_ds.read(), axis=0)
                    elif len(layersource.bands) == 1 and layersource.bands[0] == -1:
                        # If 1 band, -1 specified: dirty hack to use greyscale version of rgb image
                        image_data_tmp = image_ds.read()
                        image_data_grey = np.mean(image_data_tmp, axis=0).astype(image_data_tmp.dtype)
                        new_shape = (1, image_data_grey.shape[0], image_data_grey.shape[1])
                        image_data_grey = np.reshape(image_data_grey, new_shape)
                        if image_data_output is None:
                            image_data_output = image_data_grey
                        else:
                            image_data_output = np.append(image_data_output, image_data_grey, axis=0)
                    else:
                        # If bands specified, only read the bands to keep...
                        for band in layersource.bands:
                            # Read the band needed + reshape
                            # Remark: rasterio uses 1-based indexing instead of 0-based!!! 
                            image_data_curr = image_ds.read(band+1)
                            new_shape = (1, image_data_curr.shape[0], image_data_curr.shape[1])
                            image_data_curr = np.reshape(image_data_curr, new_shape)

                            # Set or append to image_data_output
                            if image_data_output is None:
                                image_data_output = image_data_curr
                            else:
                                image_data_output = np.append(image_data_output, image_data_curr, axis=0)

                    # Set output profile (number of bands will be corrected later on if needed) 
                    if image_profile_output is None:
                        image_profile_output = image_profile_curr

    # Write (temporary) output file
    # evade pyLance warning
    assert image_profile_output is not None
    assert isinstance(image_data_output, np.ndarray)

    # Set the number of bands to write correctly...
    if(image_format_save in [FORMAT_JPEG, FORMAT_PNG]
       and image_data_output.shape[0] == 2):
        zero_band = np.zeros(
            shape=(1, image_data_output.shape[1], image_data_output.shape[2]), 
            dtype=image_data_output.dtype)
        image_data_output = np.append(image_data_output, zero_band, axis=0)

    assert isinstance(image_data_output, np.ndarray)
    image_profile_output.update(count=image_data_output.shape[0])
    image_profile_output = _get_cleaned_write_profile(image_profile_output)
    
    # Because the (temporary) output file doesn't contain coordinates (yet),
    # suppress NotGeoreferencedWarning while writing
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=rio_errors.NotGeoreferencedWarning)

        with rio.open(str(output_filepath), 'w', **image_profile_output) as image_file:
            image_file.write(image_data_output)
    
    # If an aux.xml file was written, remove it again...
    output_aux_path = output_filepath.parent / f"{output_filepath.name}.aux.xml"
    if output_aux_path.exists() is True:
        output_aux_path.unlink()

    ##### Make the output image compliant with image_format_save #####

    # If geotiff is asked, check if the the coordinates are embedded...
    if image_format_save == FORMAT_GEOTIFF:
        # Read output image to check if coördinates are there
        with rio.open(str(output_filepath)) as image_ds:
            image_profile_orig = image_ds.profile
            image_transform_affine = image_ds.transform

            if image_pixels_ignore_border == 0:
                image_data_output = image_ds.read()
            else:
                image_data_output = image_ds.read(window=rio_windows.Window(
                        image_pixels_ignore_border, image_pixels_ignore_border, 
                        size[0], size[1]))

        logger.debug(f"original image_profile: {image_profile_orig}")

        # If coordinates are not embedded add them, if image_pixels_ignore_border
        # change them
        if((image_transform_affine[2] == 0 and image_transform_affine[5] == 0)
            or image_pixels_ignore_border > 0):
            logger.debug(f"Coordinates not present in image, driver: {image_profile_orig['driver']}")

            # If profile format is not gtiff, create new profile
            if image_profile_orig['driver'] != 'GTiff':
                image_profile_gtiff = rio_profiles.DefaultGTiffProfile.defaults

                # Copy appropriate info from source file
                image_profile_gtiff.update(
                        count=image_profile_orig['count'], width=size[0], height=size[1],
                        nodata=image_profile_orig['nodata'], dtype=image_profile_orig['dtype'])
                image_profile = image_profile_gtiff
            else:
                image_profile = image_profile_orig
                
            # Set the asked compression
            image_profile.update(compress=tiff_compress)

            # For some coordinate systems apparently the axis ordered is configured wrong in LibOWS :-(
            crs_pixel_x_size = (bbox[2]-bbox[0])/size[0]
            crs_pixel_y_size = (bbox[1]-bbox[3])/size[1]

            logger.debug(f"Coordinates to put in geotiff:\n" +
                         f"    - x-component of the pixel width, W-E: {crs_pixel_x_size}\n" +
                         f"    - y-component of the pixel width, W-E (0 if image is exactly N up): 0\n" +
                         f"    - top-left x: {bbox[0]}\n" +
                         f"    - x-component of the pixel height, N-S (0 if image is exactly N up): \n" +
                         f"    - y-component of the pixel height, N-S: {crs_pixel_y_size}\n" +
                         f"    - top-left y: {bbox[3]}")

            # Add transform and crs to the profile
            image_profile.update(
                    transform=rio_transform.Affine(
                            crs_pixel_x_size, 0, bbox[0],
                            0 , crs_pixel_y_size, bbox[3]),
                    crs=crs)

            # Delete output file, and write again
            output_filepath.unlink()
            with rio.open(str(output_filepath), 'w', **image_profile) as image_file:
                image_file.write(image_data_output)

    else:
        # For file formats that doesn't support coordinates, we add a worldfile       
        crs_pixel_x_size = (bbox[2]-bbox[0])/size[0]
        crs_pixel_y_size = (bbox[1]-bbox[3])/size[1]

        path_noext = output_filepath.parent / output_filepath.stem
        ext_world = _get_world_ext_for_image_format(image_format_save)
        output_worldfile_filepath = Path(str(path_noext) + ext_world)
        
        with output_worldfile_filepath.open('w') as wld_file:
            wld_file.write(f"{crs_pixel_x_size}")
            wld_file.write("\n0.000")
            wld_file.write("\n0.000")
            wld_file.write(f"\n{crs_pixel_y_size}")
            wld_file.write(f"\n{bbox[0]}")
            wld_file.write(f"\n{bbox[3]}")
        
        # If the image format to save is different, or if a border needs to be ignored
        if(image_format != image_format_save 
           or image_pixels_ignore_border > 0):

            # Read image 
            with rio.open(str(output_filepath)) as image_ds:
                image_profile_orig = image_ds.profile
                image_transform_affine = image_ds.transform

                # If border needs to be ignored, only read data we are interested in
                if image_pixels_ignore_border == 0:
                    image_data_output = image_ds.read()
                else:
                    image_data_output = image_ds.read(window=rio_windows.Window(
                            image_pixels_ignore_border, image_pixels_ignore_border, 
                            size[0], size[1]))
            
            # If same save format, reuse profile
            if image_format == image_format_save:
                image_profile_curr = image_profile_orig
                if image_pixels_ignore_border != 0:
                    image_profile_curr.update(width=size[0], height=size[1])
            else:
                if image_format_save == FORMAT_TIFF:
                    driver = 'GTiff'
                    compress=tiff_compress
                else:
                    raise Exception(f"Unsupported image_format_save: {image_format_save}")
                image_profile_curr = rio_profiles.Profile(
                        width=size[0], height=size[1], count=image_profile_orig['count'],
                        nodata=image_profile_orig['nodata'], dtype=image_profile_orig['dtype'],
                        compress=compress, driver=driver)

            # Delete output file, and write again
            output_filepath.unlink()
            image_profile_curr = _get_cleaned_write_profile(image_profile_curr)
            with rio.open(str(output_filepath), 'w', **image_profile_curr) as image_file:
                image_file.write(image_data_output)               

            #raise Exception(f"Different save format not supported between {image_format} and {image_format_save}")

    return output_filepath
        
def _create_filename(
        crs: pyproj.CRS,
        bbox,
        size,
        image_format: str,
        layername: str = None):
    
    # Get image extension based on format
    image_ext = _get_ext_for_image_format(image_format)

    # Use different file names for projected vs geographic crs
    if crs.is_projected:
        output_filename = f"{bbox[0]:06.0f}_{bbox[1]:06.0f}_{bbox[2]:06.0f}_{bbox[3]:06.0f}_{size[0]}_{size[1]}"
    else:
        output_filename = f"{bbox[0]:09.4f}_{bbox[1]:09.4f}_{bbox[2]:09.4f}_{bbox[3]:09.4f}_{size[0]}_{size[1]}"

    # Add layername if it is not None    
    if layername is not None:
        output_filename += "_" + layername
    
    # Add extension
    output_filename += image_ext

    return output_filename

def _get_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT
    else:
        raise Exception(f"get_ext_for_image_format for image format {image_format} is not implemented!")

def _get_world_ext_for_image_format(image_format: str) -> str:
    # Choose image extension based on format
    if image_format == FORMAT_GEOTIFF:
        return FORMAT_GEOTIFF_EXT_WORLD
    elif image_format == FORMAT_TIFF:
        return FORMAT_TIFF_EXT_WORLD
    elif image_format == FORMAT_JPEG:
        return FORMAT_JPEG_EXT_WORLD
    elif image_format == FORMAT_PNG:
        return FORMAT_PNG_EXT_WORLD
    else:
        raise Exception(f"get_world_ext_for_image_format for image format {image_format} is not implemented!")

def _get_cleaned_write_profile(profile: rio_profiles.Profile) -> Union[dict, rio_profiles.Profile]:

    # Depending on the driver, different profile keys are supported    
    if profile['driver'] == 'JPEG':
        # Don't copy profile keys to cleaned version that are not supported for JPEG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in ['tiled', 'compress', 'interleave', 'photometric']:
                profile_cleaned[profile_key] = profile[profile_key]
    elif profile['driver'] == 'PNG':
        # Don't copy profile keys to cleaned version that are not supported for JPEG
        profile_cleaned = {}
        for profile_key in profile:
            if profile_key not in ['tiled', 'interleave']:
                profile_cleaned[profile_key] = profile[profile_key]
    else:
        profile_cleaned = profile.copy()

    return profile_cleaned



# -*- coding: utf-8 -*-
"""
Modile with generic Utility functions for vector manipulations.
"""

import logging
import math
from typing import Tuple, Union

import geopandas as gpd
import pyproj
import shapely.ops as sh_ops

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------    
def calc_onborder(
        geoms_gdf: gpd.GeoDataFrame,
        border_bounds: Tuple[float, float, float, float],
        onborder_column_name: str = "onborder") -> gpd.GeoDataFrame:
    """
    Add/update a column to the GeoDataFrame with:
        * 0 if the polygon isn't on the border and 
        * 1 if it is.
        
    Args
        geoms_gdf: input GeoDataFrame
        border_bounds: the bounds (tupple with (xmin, ymin, xmax, ymax)
                       to check against to determine onborder
        onborder_column_name: the column name of the onborder column
        
    """
    # Split geoms that need unioning versus geoms that don't
    # -> They are on the edge of a tile 
    if geoms_gdf is not None and len(geoms_gdf.index) > 0:
        
        # Check 
        for i, geom_row in geoms_gdf.iterrows():
            # Check if the geom is on the border of the tile
            geom_bounds = geom_row['geometry'].bounds # type: ignore
            onborder = 0
            if(geom_bounds[0] <= border_bounds[0] 
               or geom_bounds[1] <= border_bounds[1] 
               or geom_bounds[2] >= border_bounds[2]
               or geom_bounds[3] >= border_bounds[3]):
                onborder = 1
            
            geoms_gdf.loc[i, onborder_column_name] = onborder # type: ignore

    geoms_gdf.reset_index(drop=True, inplace=True)
    return geoms_gdf

if __name__ == '__main__':
    raise Exception("Not supported")



# -*- coding: utf-8 -*-
"""
Script to load images from a WMS server.
"""

import argparse
import logging
from pathlib import Path
import shlex
import sys
import traceback

import pyproj

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.util import log_util
from orthoseg.util import ows_util

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def load_images_argstr(argstr):
    args = shlex.split(argstr)
    load_images_args(args)

def load_images_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument('-c', '--config', type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    load_images(config_path=Path(args.config))

def load_images(
        config_path: Path,
        load_testsample_images: bool = False):
    """
    Load and cache images for a segmentation project.
    
    Args:
        config_path (Path): Path to the projects config file.
        load_testsample_images (bool, optional): True to only load testsample 
            images. Defaults to False.
    """

    ##### Init #####   
    # Load the config and save in a bunch of global variables zo it 
    # is accessible everywhere 
    conf.read_orthoseg_config(config_path)
    
    # Init logging
    log_util.clean_log_dir(
            log_dir=conf.dirs.getpath('log_dir'),
            nb_logfiles_tokeep=conf.logging.getint('nb_logfiles_tokeep'))     
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath('log_dir'), __name__)      
    
    # Log + send email
    message = f"Start load_images for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")    
    email_helper.sendmail(message)
    
    try:
        # Use different setting depending if testsample or all images
        if load_testsample_images:
            output_image_dir=conf.dirs.getpath('predictsample_image_input_dir')

            # Use the same image size as for the training, that is the most 
            # convenient to check the quality
            image_pixel_width = conf.train.getint('image_pixel_width')
            image_pixel_height = conf.train.getint('image_pixel_height')
            image_pixel_x_size = conf.train.getfloat('image_pixel_x_size')
            image_pixel_y_size = conf.train.getfloat('image_pixel_y_size')
            image_pixels_overlap = 0
            image_format = ows_util.FORMAT_JPEG
            
            # To create the testsample, fetch only on every ... images
            column_start = 1
            nb_images_to_skip = 50
            
        else:
            output_image_dir=conf.dirs.getpath('predict_image_input_dir')
            
            # Get the image size for the predict
            image_pixel_width = conf.predict.getint('image_pixel_width')
            image_pixel_height = conf.predict.getint('image_pixel_height')
            image_pixel_x_size = conf.predict.getfloat('image_pixel_x_size')
            image_pixel_y_size = conf.predict.getfloat('image_pixel_y_size')
            image_pixels_overlap = conf.predict.getint('image_pixels_overlap')
            image_format = ows_util.FORMAT_JPEG
            
            # For the real prediction dataset, no skipping obviously...
            column_start = 0
            nb_images_to_skip = 0
        
        # Get ssl_verify setting
        ssl_verify = conf.general['ssl_verify']
        # Get the download cron schedule
        download_cron_schedule = conf.download['cron_schedule']

        # Get the layer info
        predict_layer = conf.predict['image_layer']
        layersources = conf.image_layers[predict_layer]['layersources']
        nb_concurrent_calls = conf.image_layers[predict_layer]['nb_concurrent_calls']
        crs = pyproj.CRS.from_user_input(conf.image_layers[predict_layer]['projection'])
        bbox = conf.image_layers[predict_layer]['bbox']
        grid_xmin = conf.image_layers[predict_layer]['grid_xmin']
        grid_ymin = conf.image_layers[predict_layer]['grid_ymin']
        image_pixels_ignore_border = conf.image_layers[predict_layer]['image_pixels_ignore_border']
        roi_filepath = conf.image_layers[predict_layer]['roi_filepath']

        # Now we are ready to get the images...
        ows_util.get_images_for_grid(
                layersources=layersources,
                output_image_dir=output_image_dir,
                crs=crs,
                image_gen_bounds=bbox,
                image_gen_roi_filepath=roi_filepath,
                grid_xmin=grid_xmin,
                grid_ymin=grid_ymin,
                image_crs_pixel_x_size=image_pixel_x_size,
                image_crs_pixel_y_size=image_pixel_y_size,
                image_pixel_width=image_pixel_width,
                image_pixel_height=image_pixel_height,
                image_pixels_ignore_border=image_pixels_ignore_border,
                nb_concurrent_calls=nb_concurrent_calls,
                cron_schedule=download_cron_schedule,
                image_format=image_format,
                pixels_overlap=image_pixels_overlap,
                column_start=column_start,
                nb_images_to_skip=nb_images_to_skip,
                ssl_verify=ssl_verify)

        # Log and send mail
        message = f"Completed load_images for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running load_images for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}")
        raise Exception(message) from ex

def main():
    try:
        load_images_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise

# If the script is ran directly...
if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the configuration of orthoseg.
"""

import configparser
import json
import logging
from pathlib import Path
import pprint

from orthoseg.util import config_util

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

# Define the chars that cannot be used in codes that are use in filenames.
# Remark: '_' cannot be used because '_' is used as devider to parse filenames, and if it is 
# used in codes as well the parsing becomes a lot more difficult. 
illegal_chars_in_codes = ['_', ',', '.', '?', ':']

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def read_orthoseg_config(config_path: Path):

    # Determine list of config files that should be loaded
    config_paths = config_util.get_config_files(config_path)
    # Load them
    global config
    config = config_util.read_config_ext(config_paths)

    # Now do orthoseg-specific checks, inits,... on config
    global config_filepaths_used
    config_filepaths_used = config_paths

    # Now set global variables to each section as shortcuts
    global general
    general = config['general']
    global model
    model = config['model'] 
    global download
    download = config['download']
    global train
    train = config['train']
    global predict
    predict = config['predict']
    global postprocess
    postprocess = config['postprocess']
    global dirs 
    dirs = config['dirs']
    global files
    files = config['files']
    global logging
    logging = config['logging']
    global email
    email = config['email']

    # Some checks to make sure the config is loaded properly
    segment_subject = general.get('segment_subject')
    if segment_subject is None or segment_subject  == 'MUST_OVERRIDE':
        raise Exception(f"Projectconfig parameter general.segment_subject needs to be overruled to a proper name in a specific project config file, \nwith config_filepaths {config_paths}")
    elif any(illegal_character in segment_subject for illegal_character in illegal_chars_in_codes):
        raise Exception(f"Projectconfig parameter general.segment_subject ({segment_subject}) should not contain any of the following characters: {illegal_chars_in_codes}")

    # If the projects_dir parameter is a relative path, resolve it towards the location of
    # the project config file.
    projects_dir = dirs.getpath('projects_dir')
    if not projects_dir.is_absolute():
        projects_dir_absolute = (config_paths[-1].parent / projects_dir).resolve()
        logger.info(f"Parameter dirs.projects_dir was relative, so is now resolved to {projects_dir_absolute}")
        dirs['projects_dir'] = projects_dir_absolute.as_posix()

    # Read the layer config
    layer_config_filepath = files.getpath('image_layers_config_filepath')
    global layer_config_filepath_used
    layer_config_filepath_used = layer_config_filepath

    global image_layers
    image_layers = read_layer_config(layer_config_filepath=layer_config_filepath)

def read_layer_config(layer_config_filepath: Path) -> dict:
    # Init
    if not layer_config_filepath.exists():
        raise Exception(f"Layer config file not found: {layer_config_filepath}")

    # Read config file...
    layer_config = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            converters={'list': lambda x: [i.strip() for i in x.split(',')],
                        'listint': lambda x: [int(i.strip()) for i in x.split(',')],
                        'dict': lambda x: None if x is None else json.loads(x),
                        'path': lambda x: Path(x)})
    layer_config.read(layer_config_filepath)
    
    # Prepare data
    image_layers = {}
    for image_layer in layer_config.sections():
        # First check if the image_layer code doesn't contain 'illegal' characters
        if any(illegal_char in image_layer for illegal_char in illegal_chars_in_codes):
            raise Exception(f"Section name [{image_layer}] in layer config should not contain any of the following characters: {illegal_chars_in_codes}, in {layer_config_filepath}")

        # Init layer with all parameters in the section as dict
        image_layers[image_layer] = dict(layer_config[image_layer])
        
        # If the layer source(s) are specified in a json parameter, parse it
        if 'layersources' in image_layers[image_layer]:
            image_layers[image_layer]['layersources'] = layer_config[image_layer].getdict('layersources')
            
            # Give default values to some optional properties of a server
            for layersource in image_layers[image_layer]['layersources']:
                if 'random_sleep' not in layersource:
                    layersource['random_sleep'] = 0
            
        else:
            # If not, the layersource should be specified in seperate parameters
            layersource = {}
            layersource['wms_server_url'] = layer_config[image_layer].get('wms_server_url')
            layersource['wms_version'] = layer_config[image_layer].get('wms_version', fallback='1.3.0')
            # The layer names and layer styles are lists
            layersource['layernames'] = layer_config[image_layer].getlist('wms_layernames')
            layersource['layerstyles'] = layer_config[image_layer].getlist('wms_layerstyles')
            # Some more properties
            layersource['bands'] = layer_config[image_layer].getlist('bands', fallback=None)
            layersource['random_sleep'] = layer_config[image_layer].getint('random_sleep', fallback=0)
            image_layers[image_layer]['layersources'] = [layersource]
        
        # Read nb_concurrent calls param
        image_layers[image_layer]['nb_concurrent_calls'] = (
                layer_config[image_layer].getint('nb_concurrent_calls', fallback=6))
            
        # Check if a region of interest is specified as file or bbox
        roi_filepath = layer_config[image_layer].getpath('roi_filepath', fallback=None)
        image_layers[image_layer]['roi_filepath'] = roi_filepath
        bbox_tuple = None
        if layer_config.has_option(image_layer, 'bbox'):
            bbox_list = layer_config[image_layer].getlist('bbox')
            bbox_tuple = (float(bbox_list[0]), float(bbox_list[1]), 
                            float(bbox_list[2]), float(bbox_list[3]))
            image_layers[image_layer]['bbox'] = bbox_tuple
        image_layers[image_layer]['bbox'] = bbox_tuple

        # Check if the grid xmin and xmax are specified            
        grid_xmin = 0
        if layer_config.has_option(image_layer, 'grid_xmin'):
            grid_xmin = layer_config[image_layer].getfloat('grid_xmin')                
        image_layers[image_layer]['grid_xmin'] = grid_xmin
        grid_ymin = 0
        if layer_config.has_option(image_layer, 'grid_ymin'):
            grid_ymin = layer_config[image_layer].getfloat('grid_ymin')
        image_layers[image_layer]['grid_ymin'] = grid_ymin

        # Check if a image_pixels_ignore_border is specified
        image_pixels_ignore_border = layer_config[image_layer].getint(
                'image_pixels_ignore_border', fallback=0)
        image_layers[image_layer]['image_pixels_ignore_border'] = image_pixels_ignore_border                                
    return image_layers

def pformat_config():
    message = f"Config files used: {pprint.pformat(config_filepaths_used)} \n"
    message += f"Layer config file used: {layer_config_filepath_used} \n"
    message += "Config info listing:\n"
    message += pprint.pformat(config_util.as_dict(config))
    message += "Layer config info listing:\n"
    message += pprint.pformat(image_layers)
    return message
    
# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Module with high-level operations to segment images.
"""

from concurrent import futures
import datetime
import json
import logging
import multiprocessing
from pathlib import Path
import shutil
import tempfile
import time
from typing import List, Optional

from geofileops import geofile
import numpy as np
import rasterio as rio
import rasterio.crs as rio_crs
import rasterio.plot as rio_plot
import tensorflow as tf
import tensorflow.keras.models as kr_models

import orthoseg.lib.postprocess_predictions as postp
from orthoseg.helpers.progress_helper import ProgressHelper
from orthoseg.util import general_util

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def predict_dir(
        model: kr_models.Model,
        input_image_dir: Path,
        output_image_dir: Path,
        output_vector_path: Optional[Path],         
        classes: list,
        prediction_cleanup_params = None,
        border_pixels_to_ignore: int = 0,
        min_pixelvalue_for_save: int = 127,
        projection_if_missing: str = None,
        input_mask_dir: Optional[Path] = None,
        batch_size: int = 16,                
        evaluate_mode: bool = False,
        cancel_filepath: Optional[Path] = None,
        force: bool = False):
    """
    Create a prediction for all the images in the directories specified 
    using the model specified.
    
    If evaluate_mode is False, the output folder(s) will contain:
        * the "raw" prediction for every image (if there are white pixels)
        * a geojson with the vectorized prediction, with a column "onborder"
          for each feature that is 1 if the feature is on the border of the
          tile, taking the border_pixels_to_ignore in account if applicable.
          This columns can be used to speed up union operations afterwards, 
          because only features on the border of tiles need to be unioned.
    
    If evaluate_mode is True, the results will all be put in the root of the 
    output folder, and the following files will be outputted:
        * the original image
        * the mask that was provided, if available
        * the "raw" prediction
        * a "cleaned" version of the prediction
    The files will in this case be prefixed with a number so they are ordered 
    in a way that is interesting for evaluation. If a mask was available, this
    prefix will be the % overlap of the mask and the prediction. If no mask is
    available, the prefix is the % white pixels in the prediction.
        
    Args
        input_image_dir (Pathlike): dir where the input images are located
        output_base_dir (Pathlike): dir where the output will be put
        output_vector_path (Pathlike): the path to write the vector output to
        classes (list): a list of the different class names. Mandatory 
            if more than background + 1 class.
        prediction_cleanup_params (dict, optional): parameters to specify which
            cleanups of the prediction need to be executed. 
            Default is None. 
        border_pixels_to_ignore: because the segmentation at the borders of the
            input images images is not as good, you can specify that x 
            pixels need to be ignored
        min_pixelvalue_for_save: the minimum pixel value that should be 
                present in the prediction to save the prediction
        input_mask_dir: optional dir where the mask images are located
        projection_if_missing: Normally the projection should be in the raster file. If it 
                    is not, you can explicitly specify one.
        batch_size: batch size to use while predicting. This must be choosen 
                depending on the neural network architecture and available 
                memory on you GPU.
        evaluate_mode: True to run in evaluate mode
        cancel_filepath: If the file in this path exists, processing stops asap
        force: False to skip images that already have a prediction, true to
               ignore existing predictions and overwrite them
    """
    
    ### Init ###
    if not input_image_dir.exists():
        logger.warn(f"In predict_dir, but input_image_dir doesn't exist, so return: {input_image_dir}")
        return
    if output_vector_path is not None and output_vector_path.exists():
        logger.warn(f"In predict_dir, output file exists already, so return: {output_vector_path}")
        return
    tmp_dir = Path(tempfile.gettempdir()) / Path(__file__).stem

    logger.info(f"Start predict for input_image_dir: {input_image_dir}")

    # Eager and not eager prediction seems +- the same performance-wise
    #model.run_eagerly = False
    
    # If we are using evaluate mode, change the output dir...
    if evaluate_mode:
        output_image_dir = Path(str(output_image_dir) + '_eval')

    # Create the output dir's if they don't exist yet...
    for dir in [output_image_dir, tmp_dir]:
        if not dir.exists():
            dir.mkdir()

    # Write prediction config used, so it can be used for postprocessing
    prediction_config_path = output_image_dir / "prediction_config.json"
    with open(prediction_config_path, 'w') as pred_conf_file:
        pred_conf = {}
        pred_conf['border_pixels_to_ignore'] = border_pixels_to_ignore
        pred_conf['classes'] = classes
        json.dump(pred_conf, pred_conf_file)

    ### Get list of all image files to process and to skip ###
    image_filepaths: List[Path] = []
    input_ext = ['.png', '.tif', '.jpg']
    for input_ext_cur in input_ext:
        image_filepaths.extend(input_image_dir.rglob('*' + input_ext_cur))
    image_filepaths = sorted(image_filepaths)
    nb_images = len(image_filepaths)
    logger.info(f"Found {nb_images} {input_ext} images to predict on in {input_image_dir}")
    
    # If force is false, get list of all existing predictions
    # Getting the list once is way faster than checking file per file later on!
    images_done_log_filepath = output_image_dir / 'images_done.txt'
    images_error_log_filepath = output_image_dir / 'images_error.txt'
    image_done_filenames = set()
    if force is False:
        # First read the listing files if they exists
        if images_done_log_filepath.exists():
            with images_done_log_filepath.open() as f:
                for filename in f:
                    image_done_filenames.add(filename.rstrip())
        if images_error_log_filepath.exists():                    
            with images_error_log_filepath.open() as f:
                for filename in f:
                    image_done_filenames.add(filename.rstrip())
        if len(image_done_filenames) > 0:
            logger.info(f"Found {len(image_done_filenames)} predicted images in {images_done_log_filepath} and {images_error_log_filepath}, they will be skipped")

    # First write output to tmp files, so we know if the process completed correctly or not...
    pred_tmp_output_path = None
    if output_vector_path is not None:
        pred_tmp_output_path = output_image_dir / f"{output_vector_path.stem}_tmp.gpkg"
        pred_tmp_output_lock_path = Path(f"{str(pred_tmp_output_path)}.lock")
        # if lock file exists, remove it:
        if pred_tmp_output_lock_path.exists():
            pred_tmp_output_lock_path.unlink()

    ### Loop through all files to process them ###
    nb_parallel_read = batch_size*6
    nb_parallel_postprocess = multiprocessing.cpu_count()
    predict_images = []
    nb_to_process = nb_images
    nb_processed = 0
    progress = None
    postp_future_to_input_path = {}
    read_future_to_input_path = {}
    image_id = -1
    last_image_reached = False

    # We don't want the postprocess workers to block the entire system, 
    # so make them a bit nicer 
    def init_postprocess_worker():
        general_util.setprocessnice(15)
    
    with futures.ThreadPoolExecutor(nb_parallel_read) as read_pool, \
         futures.ProcessPoolExecutor(
                nb_parallel_postprocess, 
                initializer=init_postprocess_worker()) as postprocess_pool:

        # Start looping. 
        # If ready to stop, the code below will break
        perf_time_start = datetime.datetime.now()
        while True:
            
            # If we are ready, stop!
            if(last_image_reached is True
                    and len(read_future_to_input_path) == 0
                    and len(postp_future_to_input_path) == 0):
                break

            # If the cancel file exists, stop processing...
            if cancel_filepath is not None and cancel_filepath.exists():
                print()
                logger.info(f"Cancel file found, so stop: {cancel_filepath}")
                break
            
            # Get the next filepath to be processed
            image_filepath = None
            if image_id < (len(image_filepaths)-1):
                image_id += 1
                image_filepath = image_filepaths[image_id]
                
                # Check if the image has been processed already
                if force is False and image_filepath.name in image_done_filenames:
                    logger.debug(f"Predict for image has already been done before and force is False, so skip: {image_filepath.name}")
                    nb_to_process -= 1
                    continue
            
                nb_processed += 1
            
                # Schedule file to be read 
                read_future = read_pool.submit(
                        read_image,               # Function 
                        image_filepath,           # Arg 1
                        projection_if_missing)    # Arg 2)
                read_future_to_input_path[read_future] = image_filepath
            else:
                last_image_reached = True
            
            # Prepare batch_size images that have been read for prediction
            read_sleep_logged = False
            while len(predict_images) < batch_size:

                # If we are at the last image and all images are read, break
                # so the final batch can be predicted
                if last_image_reached is True and len(read_future_to_input_path) == 0:
                    break
                # If the read pool is not full, first schedule extra reads
                if(last_image_reached is False 
                        and len(read_future_to_input_path) < nb_parallel_read):
                    break

                # Prepare the images that have been read for predicting 
                futures_done = [future for future in read_future_to_input_path if future.done() is True]
                for future in futures_done:
                
                    # Prediction can only handle batch_size images
                    if len(predict_images) >= batch_size:
                        break

                    try:
                        # Get the result from the read
                        read_result = future.result()
                        image_filepath_read = read_result["image_filepath"]
                        
                        # Prepare the filepath for the output
                        output_suffix = '.tif'
                        if evaluate_mode:
                            # In evaluate mode, put everyting in output base dir for easier 
                            # comparison
                            output_image_pred_dir = output_image_dir

                            # Prepare complete filepath for image prediction
                            output_image_pred_path = output_image_dir / image_filepath_read.stem
                        else:
                            # If saving predictions to images for real, keep hierarchic structure if present
                            tmp_output_filepath = Path(str(image_filepath_read).replace(str(input_image_dir), str(output_image_dir)))
                            output_image_pred_dir = tmp_output_filepath.parent
                            output_image_pred_path = output_image_pred_dir / f"{image_filepath_read.stem}_pred{output_suffix}"
                        
                        predict_images.append({
                                "input_image_filepath": image_filepath_read,
                                "output_pred_filepath": output_image_pred_path,
                                "output_image_pred_dir": output_image_pred_dir,
                                "image_crs": read_result["image_crs"],
                                "image_transform": read_result["image_transform"],
                                "image_data": read_result["image_data"]})

                    finally:
                        # Remove from queue...
                        del read_future_to_input_path[future]

                # If not at last image + not enough images yet for predict + read 
                # queue is full, sleep
                if(last_image_reached is False 
                        and len(predict_images) < batch_size 
                        and len(read_future_to_input_path) >= nb_parallel_read):
                    if read_sleep_logged is False:
                        logger.info("Wait for images to be read")
                        read_sleep_logged = True
                    time.sleep(0.01)

            # If batch_size images are ready for prediction or we are at 
            # the last images: predict
            if len(predict_images) == batch_size or last_image_reached is True:
                
                perf_time_now = datetime.datetime.now()
                perfinfo = f"waiting for read took {perf_time_now-perf_time_start}"
                perf_time_start = perf_time_now

                # Init progress only at 2nd batch, as the first is very slow
                if progress is None and nb_processed > batch_size*2:
                    progress = ProgressHelper(
                            message=f"predict to {output_image_dir.parent.name}/{output_image_dir.name}",
                            nb_steps_total=nb_to_process, 
                            nb_steps_done=batch_size)
                
                ## Predict! ##
                logger.debug(f"Start prediction for {len(predict_images)} images")
                perf_time_start = datetime.datetime.now()
                curr_batch_image_list = [curr_batch_image_info['image_data'] for curr_batch_image_info in predict_images]
                curr_batch_image_arr = np.stack(curr_batch_image_list)
                curr_batch_image_pred_arr = model.predict_on_batch(curr_batch_image_arr)
                
                perf_time_now = datetime.datetime.now()
                perfinfo += f", predict took {perf_time_now-perf_time_start}"
                perf_time_start = perf_time_now

                # In tf > 2.1 a tf.tensor object is returned, but we want an ndarray
                if type(curr_batch_image_pred_arr) is tf.Tensor:
                    curr_batch_image_pred_arr = np.array(curr_batch_image_pred_arr.numpy())
                else:
                    curr_batch_image_pred_arr = np.array(curr_batch_image_pred_arr)
                
                ## Save predictions ##
                # Remark: trying to parallelize this doesn't seem to help at all!
                logger.debug("Start post-processing")    
                for batch_image_id, image_info in enumerate(predict_images):
                    try:
                        # If not in evaluate mode... save to vector in background
                        if evaluate_mode is False and output_vector_path is not None:
                            # Prepare prediction array...
                            #   - 1st channel is background, don't postprocess
                            #   - convert to uint8 to reduce pickle size/time    
                            image_pred_arr_uint8 = ((curr_batch_image_pred_arr[batch_image_id,:,:,1:]) * 255).astype(np.uint8)
                            future = postprocess_pool.submit(
                                    postp.polygonize_pred_multiclass_to_file,
                                    image_pred_arr_uint8,
                                    image_info['image_crs'],
                                    image_info['image_transform'],
                                    min_pixelvalue_for_save,
                                    classes[1:],               # The first class is the background so skip that
                                    pred_tmp_output_path,
                                    prediction_cleanup_params,
                                    border_pixels_to_ignore)
                            postp_future_to_input_path[future] = image_info['input_image_filepath']
                        
                        else:
                            # Saving the predictions as images at the moment only used 
                            # for evaluate mode...
                            # TODO: would ideally be moved to the background 
                            # processing as well to simplify code here...
                            postp.clean_and_save_prediction(
                                    image_image_filepath=image_info['input_image_filepath'],
                                    image_crs=image_info['image_crs'],
                                    image_transform=image_info['image_transform'],
                                    image_pred_arr=curr_batch_image_pred_arr[batch_image_id],
                                    output_dir=image_info['output_image_pred_dir'],
                                    input_image_dir=input_image_dir,
                                    input_mask_dir=input_mask_dir,
                                    border_pixels_to_ignore=border_pixels_to_ignore,
                                    min_pixelvalue_for_save=min_pixelvalue_for_save,
                                    evaluate_mode=evaluate_mode,
                                    classes=classes,
                                    force=force)
                            
                            # Write filepath to file with files that are done
                            with images_done_log_filepath.open('a+') as image_donelog_file:
                                image_donelog_file.write(f"{image_info['input_image_filepath'].name}\n")
                    except:
                        logger.exception(f"Error postprocessing pred for {image_info['input_image_filepath']}")
    
                        # Write filepath to file with errors
                        with images_error_log_filepath.open('a+') as image_errorlog_file:
                            image_errorlog_file.write(image_info['input_image_filepath'].name + '\n')
                
                perf_time_now = datetime.datetime.now()
                perfinfo += f", scheduling postprocessings took {perf_time_now-perf_time_start}"
                perf_time_start = perf_time_now

                # Poll for completed postprocessings 
                postp_sleep_logged = False 
                while len(postp_future_to_input_path) > 0:
                    
                    # If not at last file, get results from all futures that are 
                    # done, if at last file, wait till all are done
                    if last_image_reached is False:
                        futures_done = [future for future in postp_future_to_input_path if future.done() is True]
                    else:
                        logger.info("Wait for last batch")
                        futures_done = futures.wait(postp_future_to_input_path).done
                    for future in futures_done:
                        # Get the result from the polygonization
                        try:
                            # Get the result (= exception when something went wrong)
                            future.result()

                            # Write filepath to file with files that are done
                            with images_done_log_filepath.open('a+') as image_donelog_file:
                                image_donelog_file.write(postp_future_to_input_path[future].name + '\n')
                        except Exception as ex:
                            # Write filepath to file with errors
                            logger.exception(f"Error postprocessing result for {postp_future_to_input_path[future].name}")
                            with images_error_log_filepath.open('a+') as image_errorlog_file:
                                image_errorlog_file.write(postp_future_to_input_path[future].name + '\n')
                        finally:
                            # Remove from queue...
                            del postp_future_to_input_path[future]

                    # Wait till number below thresshold to evade huge waiting 
                    # list (and memory issues)
                    if len(postp_future_to_input_path) > nb_parallel_postprocess*2:
                        if postp_sleep_logged is False:
                            logger.info(f"Postprocessing seems to take longer than prediction, so wait for it to catch up")
                            postp_sleep_logged = True
                        time.sleep(0.01)
                    else:
                        # No need to wait (anymore)...
                        if postp_sleep_logged is True:
                            logger.info(f"Waited enough for postprocessing to catch up, so continue predicting") 
                        
                        perf_time_now = datetime.datetime.now()
                        perfinfo += f", after postproces: took {perf_time_now-perf_time_start}"
                        perf_time_start = perf_time_now
                        break
                
                # Reset variable for next batch
                predict_images = []

                ## Log the progress and prediction speed ##
                if len(perfinfo) > 0: 
                    logger.debug(perfinfo) 
                if progress is not None:
                    progress.step(nb_steps=batch_size)

        # If alle images were processed, rename to real output file + cleanup
        if(last_image_reached is True 
                and output_vector_path is not None 
                and pred_tmp_output_path is not None 
                and pred_tmp_output_path.exists()):
            geofile.move(pred_tmp_output_path, output_vector_path)
            geofile.rename_layer(output_vector_path, output_vector_path.stem)
            shutil.rmtree(output_image_dir)

def read_image(
        image_filepath: Path,
        projection_if_missing: str = None) -> dict:

    # Read input file
    # Because sometimes a read seems to fail, retry up to 3 times...
    retry_count = 0
    while True:
        try:
            with rio.open(str(image_filepath)) as image_ds:
                # Read geo info
                image_crs = image_ds.profile['crs']
                image_transform = image_ds.transform
                
                # Read pixels + change from (channels, width, height) to 
                # (width, height, channels) + normalize to between 0 and 1
                image_data = image_ds.read()
                image_data = rio_plot.reshape_as_image(image_data)
                image_data = image_data / 255.0 
            
            # Read worked, so jump out of the loop...
            break
        except Exception as ex:
            retry_count += 1
            logger.warning(f"Read failed, retry nb {retry_count} for {image_filepath}")
            if retry_count >= 3:
                message = f"STOP: Read failed {retry_count} times for {image_filepath}, with exception: {ex}"
                logger.error(message)
                raise Exception(message)
        
    # The read was successfull, now check if there was a projection in the 
    # file and/or if one was provided
    if image_crs is None:
        if projection_if_missing is not None:
            image_crs = rio_crs.CRS.from_string(projection_if_missing)
        else:
            message = f"Image doesn't contain projection and no projection_if_missing provided!: {image_filepath}"
            logger.error(message)
            raise Exception(message)
        
    # Now return the result
    result = {'image_data': image_data,
              'image_crs': image_crs,
              'image_transform': image_transform,
              'image_filepath': image_filepath}
    return result
        
# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage progress reporting.
"""

import datetime
import logging

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

class ProgressHelper:
    first_reporting_done = False

    def __init__(
            self, 
            message: str,
            nb_steps_total: int,
            nb_steps_done: int = 0,
            start_time: datetime.datetime = None,
            time_between_reporting_s: int = 60,
            calculate_eta_since_lastreporting: bool = True):
        self.message = message
        if start_time is None:
            self.start_time = datetime.datetime.now()
        else:
            self.start_time = start_time
        self.start_time_lastreporting = self.start_time
        self.nb_steps_total = nb_steps_total
        self.nb_steps_done = nb_steps_done
        self.nb_steps_done_lastreporting = nb_steps_done
        self.time_between_reporting_s = time_between_reporting_s
        self.calculate_eta_since_lastreporting = calculate_eta_since_lastreporting

    def step(self,
            message: str = None, 
            nb_steps: int = 1):

        # Increase done counter
        self.nb_steps_done += nb_steps
        
        # Calculate time since last reporting
        time_now = datetime.datetime.now()
        time_passed_lastprogress_s = (time_now-self.start_time_lastreporting).total_seconds()
    
        # Calculate the time_passed and nb_done we want to use to calculate ETA
        if self.calculate_eta_since_lastreporting is True:
            time_passed_for_eta_s = (time_now-self.start_time_lastreporting).total_seconds()
            nb_steps_done_eta = self.nb_steps_done - self.nb_steps_done_lastreporting
        else:
            time_passed_for_eta_s = (time_now-self.start_time).total_seconds()
            nb_steps_done_eta = self.nb_steps_done
        
        # Print progress on first step or if time between reporting has passed
        if(time_passed_lastprogress_s >= self.time_between_reporting_s
                or self.first_reporting_done is False):
            # Evade divisions by zero
            if time_passed_for_eta_s == 0 or nb_steps_done_eta == 0:
                return
                
            nb_per_hour = (nb_steps_done_eta/time_passed_for_eta_s) * 3600
            hours_to_go = (int)((self.nb_steps_total-self.nb_steps_done)/nb_per_hour)
            min_to_go = (int)((((self.nb_steps_total-self.nb_steps_done)/nb_per_hour)%1)*60)
            if message is not None:
                progress_message = f"{message}, {self.nb_steps_done}/{self.nb_steps_total}, {hours_to_go:3d}:{min_to_go:2d} left at {nb_per_hour:0.0f}/h"
            else:
                progress_message = f"{self.message}, {self.nb_steps_done}/{self.nb_steps_total}, {hours_to_go:3d}:{min_to_go:2d} left at {nb_per_hour:0.0f}/h"
            logger.info(progress_message)

            self.start_time_lastreporting = time_now
            self.nb_steps_done_lastreporting = self.nb_steps_done
            if self.first_reporting_done is False:
                self.first_reporting_done = True

# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('version.txt', mode='r') as file:
    version = file.readline()

setuptools.setup(
    name="orthoseg", 
    version=version,
    author="Pieter Roggemans",
    author_email="pieter.roggemans@gmail.com",
    description="Package to make it easier to segment orthophotos.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/theroggy/orthoseg",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
            "fiona", "geofileops>=0.3,<0.4", "geopandas", 
            "owslib", "pillow", "pycron", 
            "rasterio", "segmentation-models>=1.0,<1.1", 
            "tensorflow>=2.5,<2.9"],
    entry_points='''
            [console_scripts]
            orthoseg_load_images=orthoseg.load_images:main
            orthoseg_train=orthoseg.train:main
            orthoseg_predict=orthoseg.predict:main
            orthoseg_postprocess=orthoseg.postprocess:main
            scriptrunner=orthoseg.scriptrunner:main
            ''',
    classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)


# -*- coding: utf-8 -*-
"""
Run the scripts in a directory.
"""

import argparse
import configparser
from pathlib import Path
import subprocess
import sys
import time

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.util import log_util
from orthoseg.util import config_util

runner_config = None

def main():
    
    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument('-d', '--script_dir',
            help='Directory containing the scripts to run.')
    optional.add_argument('-w', '--watch', action='store_true', default=False,
            help='Watch the directory forever for files getting in it.')
    optional.add_argument('-c', '--config',
            help='Path to a config file with parameters that need to overrule the defaults.')
    
    # Add back help         
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')
    args = parser.parse_args()
    
    ### Init stuff ###
    script_dir = Path(args.script_dir)
    if not script_dir.exists():
        raise Exception(f"script dir {script_dir} does not exist")

    # Load the scriptrunner config
    conf = load_scriptrunner_config(args.config, script_dir)

    # Init logging
    log_util.clean_log_dir(
            log_dir=conf['dirs'].getpath('log_dir'),
            nb_logfiles_tokeep=conf['logging'].getint('nb_logfiles_tokeep'))
    logger = log_util.init_logging_dictConfig(
            logconfig_dict=conf['logging'].getdict('logconfig'),
            log_basedir=conf['dirs'].getpath('script_dir'),
            loggername=__name__)
    
    # Init working dirs
    done_dir = conf['dirs'].getpath('done_dir')
    error_dir = conf['dirs'].getpath('error_dir')

    # Loop over scripts to be ran
    wait_message_printed = False
    while True:

        # List the scripts in the dir
        script_paths = []
        script_patterns = conf['general'].getlist('script_patterns')
        for script_pattern in script_patterns:
            script_paths.extend(list(script_dir.glob(script_pattern)))

        # If no scripts found, sleep or stop...
        if len(script_paths) == 0:
            if args.watch is False:
                logger.info(f"No scripts found (anymore) in {script_dir}, so stop")
                break
            else:
                if wait_message_printed is False:
                    logger.info(f"No scripts to run in {script_dir}, so watch script dir...")
                    wait_message_printed = True
                time.sleep(10)
                continue

        # Get next script alphabetically
        script_path = sorted(script_paths)[0]
        
        try:
            # Run the script and print output in realtime
            wait_message_printed = False
            logger.info(f"Run script {script_path}")
            cmd = [script_path]

            process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, encoding='utf-8')
                    #,creationflags=subprocess.CREATE_NO_WINDOW)
            
            still_running = True
            while still_running:
                if process.stdout is not None:
                    output = process.stdout.readline()
                    if output == '' and process.poll() is not None:
                        still_running = False
                    if output:
                        logger.info(output.strip())
            
            # If error code != 0, an error occured
            rc = process.poll()
            if rc != 0:
                # Script gave an error, so move to error dir
                logger.error(f"Script {script_path} gave error return code: {rc}")
                error_dir.mkdir(parents=True, exist_ok=True)
                error_path = error_dir / script_path.name
                if error_path.exists():
                    error_path.unlink()
                script_path.rename(target=error_path)
            else:
                # Move the script to the done dir 
                done_dir.mkdir(parents=True, exist_ok=True)
                done_path = done_dir / script_path.name
                if done_path.exists():
                    done_path.unlink()
                script_path.rename(target=done_path)

        except Exception as ex:
            logger.exception(f"Error running script {script_path}")

            # If the script still exists, move it to error dir
            if script_path.exists():
                error_dir.mkdir(parents=True, exist_ok=True)
                error_path = error_dir / script_path.name
                if error_path.exists():
                    error_path.unlink()
                script_path.rename(target=error_path)

def load_scriptrunner_config(
        config_path: str,
        script_dir: Path) -> configparser.ConfigParser:

    # Load defaults first
    scriptrunner_py_dir = Path(__file__).resolve().parent
    config_paths = [scriptrunner_py_dir / 'scriptrunner_defaults.ini']

    # If a config path is specified, this config should overrule the defaults
    if config_path is not None:
        config_paths.append(Path(config_path))

    # Load!
    scriptrunner_config = config_util.read_config_ext(config_paths)

    # If a script_dir is specified, it overrides the config file(s)
    if script_dir is not None:
        scriptrunner_config['dirs']['script_dir'] = script_dir.as_posix()

    return scriptrunner_config

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Module containing some general utilities.
"""

import datetime
import logging
import os

import psutil

################################################################################
# Some init
################################################################################

logger = logging.getLogger(__name__)

class MissingRuntimeDependencyError(Exception):
    """
    Exception raised when an unsupported SQL statement is passed.

    Attributes:
        message (str): Exception message
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

################################################################################
# The real work
################################################################################

def report_progress(
        start_time: datetime.datetime,
        nb_done: int,
        nb_todo: int,
        operation: str = None,
        nb_parallel: int = 1):

    # Init
    time_passed = (datetime.datetime.now()-start_time).total_seconds()
    pct_progress = 100.0-(nb_todo-nb_done)*100/nb_todo
    
    # If we haven't really started yet, don't report time estimate yet
    if nb_done == 0:
        print(f"\r  ?: ? left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    ", 
              end="", flush=True)
    elif time_passed > 0:
        # Else, report progress properly...
        processed_per_hour = (nb_done/time_passed) * 3600
        # Correct the nb processed per hour if running parallel 
        if nb_done < nb_parallel:
            processed_per_hour = round(processed_per_hour * nb_parallel / nb_done)
        hours_to_go = (int)((nb_todo - nb_done)/processed_per_hour)
        min_to_go = (int)((((nb_todo - nb_done)/processed_per_hour)%1)*60)
        pct_progress = 100.0-(nb_todo-nb_done)*100/nb_todo
        if pct_progress < 100:
            print(f"\r{hours_to_go:3d}:{min_to_go:2d} left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    ", 
                  end="", flush=True)
        else:
            print(f"\r{hours_to_go:3d}:{min_to_go:2d} left to do {operation} on {(nb_todo-nb_done):8d} of {nb_todo:8d} ({pct_progress:3.2f}%)    \n", 
                  end="", flush=True)

def formatbytes(bytes: float):
    """
    Return the given bytes as a human friendly KB, MB, GB, or TB string
    """

    bytes_float = float(bytes)
    KB = float(1024)
    MB = float(KB ** 2) # 1,048,576
    GB = float(KB ** 3) # 1,073,741,824
    TB = float(KB ** 4) # 1,099,511,627,776

    if bytes_float < KB:
        return '{0} {1}'.format(bytes_float, 'Bytes' if bytes_float > 1 else 'Byte')
    elif KB <= bytes_float < MB:
        return '{0:.2f} KB'.format(bytes_float/KB)
    elif MB <= bytes_float < GB:
        return '{0:.2f} MB'.format(bytes_float/MB)
    elif GB <= bytes_float < TB:
        return '{0:.2f} GB'.format(bytes_float/GB)
    elif TB <= bytes_float:
        return '{0:.2f} TB'.format(bytes_float/TB)

def process_nice_to_priority_class(nice_value: int) -> int:
    if nice_value <= -15:
        return psutil.REALTIME_PRIORITY_CLASS
    elif nice_value <= -10:
        return psutil.HIGH_PRIORITY_CLASS
    elif nice_value <= -5:
        return psutil.ABOVE_NORMAL_PRIORITY_CLASS
    elif nice_value <= 0:
        return psutil.NORMAL_PRIORITY_CLASS
    elif nice_value <= 10:
        return psutil.BELOW_NORMAL_PRIORITY_CLASS
    else:
        return psutil.IDLE_PRIORITY_CLASS

def setprocessnice(nice_value: int):
    p = psutil.Process(os.getpid())
    if os.name == 'nt':
        p.nice(process_nice_to_priority_class(nice_value))
    else:
        p.nice(nice_value)

def getprocessnice() -> int:
    p = psutil.Process(os.getpid())
    nice_value = p.nice()
    if os.name == 'nt':
        if nice_value == psutil.REALTIME_PRIORITY_CLASS:
            return -20
        elif nice_value == psutil.HIGH_PRIORITY_CLASS:
            return -10
        elif nice_value == psutil.ABOVE_NORMAL_PRIORITY_CLASS:
            return -5
        elif nice_value == psutil.NORMAL_PRIORITY_CLASS:
            return 0
        elif nice_value == psutil.BELOW_NORMAL_PRIORITY_CLASS:
            return 10
        elif nice_value == psutil.IDLE_PRIORITY_CLASS:
            return 20
        else:
            return 0
    else:
        return int(nice_value)
